"""Test functions for utils scripts"""
import geopandas as gpd
import pandas as pd
import pytest
from numpy.testing import assert_almost_equal
from shapely import LineString

from pywatemsedem.tools import (
    extract_tags_from_template_file,
    format_forced_routing,
    reformat_LineString_to_source_targetf,
)


@pytest.mark.parametrize(
    "template,valid",
    [
        ("perceelskaart_2018_molenbeek_s1.rst", True),
        ("perceelskaart_2018_molenbeek_s1.rdc", True),
        ("perceelskaart_2018_molenbeek_s1.rst.aux", True),
        ("perceelskaart_2018_molenbeek_s1.rst.aux.rst", True),
        ("perceelskaart_2018_molenbeek_velm_s1.rst", False),
    ],
)
def test_extract_tags_from_template_file(template, valid):
    """Test function for extracting tags (catchment, year, scenario identifier) from a
    template file"""
    catchment, scenario, year, valid_ = extract_tags_from_template_file(template)

    assert catchment == "molenbeek"
    assert scenario == "1"
    assert year == 2018
    assert valid == valid_


class TestReformatRouting:
    """Test class for reformat LineString to Source/target dataframe"""

    coor = [(1, 1), (1, 2), (2, 1), (4, 8), (5, 6)]
    ls = LineString(coor)
    res = 1
    minmax = [0, 0, 10, 10]

    def test_trajectory_routing(self):
        """Test trajectory routing, see function for definition"""
        exp = pd.DataFrame(
            data={
                "fromX": [1, 1, 2, 4],
                "fromY": [1, 2, 1, 8],
                "toX": [1, 2, 4, 5],
                "toY": [2, 1, 8, 6],
            }
        )

        out = reformat_LineString_to_source_targetf(self.coor, trajectory_routing=True)

        assert_almost_equal(out["fromX"].values, exp["fromX"].values)
        assert_almost_equal(out["toX"].values, exp["toX"].values)
        assert_almost_equal(out["fromY"].values, exp["fromY"].values)
        assert_almost_equal(out["toY"].values, exp["toY"].values)

    def test_jump_routing(self):
        """Test jump routing, see function for definition"""
        exp = pd.DataFrame(data={"fromX": [1], "fromY": [1], "toX": [5], "toY": [6]})

        out = reformat_LineString_to_source_targetf(self.coor, trajectory_routing=False)

        assert_almost_equal(out["fromX"].values, exp["fromX"].values)
        assert_almost_equal(out["toX"].values, exp["toX"].values)
        assert_almost_equal(out["fromY"].values, exp["fromY"].values)
        assert_almost_equal(out["toY"].values, exp["toY"].values)

    def format_forced_routing(self):
        """Test reformat forced routing

        Test it with standard jump trajectory and assert output column and row
        """
        out = format_forced_routing(
            gpd.GeoDataFrame([1], geometry=[self.ls]),
            resolution=self.res,
            minmax=self.minmax,
        )

        assert_almost_equal(out["fromcol"].values, 2)
        assert_almost_equal(out["fromrow"].values, 10)
        assert_almost_equal(out["tocol"].values, 6)
        assert_almost_equal(out["torow"].values, 5)



from pathlib import Path

import geopandas as gpd
import numpy as np
import pytest
from conftest import catchment_data

from pywatemsedem.catchment import Catchment
from pywatemsedem.errors import (
    PywatemsedemRasterValueError,
    PywatemsedemVectorAttributeValueError,
)
from pywatemsedem.geo.utils import load_raster, write_arr_as_rst


class TestCatchment:
    """Test Catchment properties

    Test main functionalities for Catchment class.

    Notes
    -----
    All tests in this suite do implicit testing:
    - Does the setting of a property run?
    - Are the unique values and counts equal?
    - Are data types equal?
    """

    name = "langegracht"
    catchment = Catchment(
        name, catchment_data.catchment, catchment_data.dtm, 20, 31370, -9999
    )

    def test_kfactor(self):
        """Test assignment kfactor raster"""
        # test assignment raster
        self.catchment.kfactor = catchment_data.k

        # test type
        assert self.catchment.kfactor.arr.dtype == np.int32

        # test unique values
        un, counts = np.unique(self.catchment.kfactor.arr, return_counts=True)
        np.testing.assert_allclose(un, [-9999, 20, 27, 40, 42])
        np.testing.assert_allclose(counts, [28236, 1377, 72, 818, 18941])

    def test_kfactor_warning(self, recwarn, tmp_path):
        """Test warning wrong value kfactor raster"""
        tmp_path.mkdir(exist_ok=True)
        fname = Path(tmp_path) / "temp.rst"
        arr, profile = load_raster(catchment_data.k)
        arr[arr < 40] = -10
        write_arr_as_rst(arr, fname, arr.dtype, profile)
        self.catchment.kfactor = fname
        w = recwarn.pop(UserWarning)
        assert (
            str(w.message) == "Negative values detected in K-factor raster, "
            "setting negative values to 0."
        )

    def test_landuse(self):
        """Test assignment landuse raster"""
        self.catchment.landuse = catchment_data.basemap
        print(np.unique(self.catchment.landuse.arr, return_counts=True))

    def test_landuse_wrong_value(self, tmp_path):
        """Test wrong value in landuse raster"""
        fname = Path(tmp_path) / "temp.rst"
        arr, profile = load_raster(catchment_data.basemap)
        arr[arr == 10] = -7
        write_arr_as_rst(arr, fname, arr.dtype, profile)
        with pytest.raises(
            PywatemsedemRasterValueError, match="can only contain values"
        ):
            self.catchment.landuse = fname

    @pytest.mark.saga
    def test_river(self):
        """Test assignment river raster and derivative segments and routing rasters"""
        # test assign vector
        self.catchment.vct_river = catchment_data.river

        # test raster river type and unique values
        un, counts = np.unique(self.catchment.river.arr, return_counts=True)
        np.testing.assert_allclose(un, np.array([-9999, -1]))
        np.testing.assert_allclose(counts, np.array([48910, 534]))

        # test raster routing type and unique values
        un, counts = np.unique(self.catchment.routing.arr, return_counts=True)
        np.testing.assert_allclose(un, np.array([0, 1, 3, 5, 7]))
        np.testing.assert_allclose(counts, np.array([48911, 84, 257, 152, 40]))

        # test segments
        un, counts = np.unique(self.catchment.segments.arr, return_counts=True)
        np.testing.assert_allclose(
            un,
            np.array([0, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18]),
        )
        np.testing.assert_allclose(
            counts,
            np.array(
                [48910, 11, 5, 4, 46, 60, 3, 14, 27, 5, 41, 43, 28, 87, 43, 52, 23, 42]
            ),
        )

    def test_tubed_river(self):
        """Test tubed river and error"""
        gdf = gpd.read_file(catchment_data.river)

        tubed = gdf.iloc[0:2]
        non_tubed = gdf.iloc[2:]

        # check if functionality runs
        self.catchment.vct_river = non_tubed
        self.catchment.vct_tubed_river = tubed

        # check for overlap
        self.catchment.vct_river = gdf
        with pytest.raises(IOError, match="are also present"):
            self.catchment.vct_tubed_river = tubed

    def test_infrastructure(self):
        """Test assignment infrastructure vector"""
        # test assign vector
        self.catchment.vct_infrastructure_buildings = catchment_data.infrastructure
        self.catchment.vct_infrastructure_roads = catchment_data.roads

        # test infrastructure river type and unique values
        un, counts = np.unique(
            self.catchment.infrastructure_roads.arr, return_counts=True
        )
        np.testing.assert_allclose(un, [-9999, -7.0, -2])
        np.testing.assert_allclose(counts, [46746, 1408, 1290])

        un, counts = np.unique(
            self.catchment.infrastructure_buildings.arr, return_counts=True
        )
        np.testing.assert_allclose(un, [-9999, -2])
        np.testing.assert_allclose(counts, [46834, 2610])

        un, counts = np.unique(self.catchment.infrastructure.arr, return_counts=True)
        np.testing.assert_allclose(un, [-9999, -7.0, -2])
        np.testing.assert_allclose(counts, [45353, 1408, 2683])

    def test_infrastructure_wrong_value(self):
        "Test wrong value in 'paved' attribute infr roads"
        df = gpd.read_file(catchment_data.roads)
        df["paved"] = -3
        with pytest.raises(
            PywatemsedemVectorAttributeValueError,
            match="vector can only contain values",
        ):
            self.catchment.vct_infrastructure_roads = df



# Standard libraries
import logging
import os
import subprocess
import tempfile
from copy import deepcopy
from functools import wraps
from pathlib import Path

import fiona
import geopandas as gpd
import numpy as np
import pandas as pd
import rasterio
from rasterio.features import shapes
from rasterio.merge import merge

from pywatemsedem.geo.rasterproperties import RasterProperties

from ..defaults import (
    SAGA_FLAGS,
    SUFFIXES_RST,
    SUFFIXES_SAGA,
    SUFFIXES_SHP,
    SUFFIXES_TIF,
    SUFFIXES_TXT,
)
from .valid import (
    valid_input,
    valid_linesvector,
    valid_pointvector,
    valid_polygonvector,
    valid_raster,
    valid_rasterlist,
    valid_vector,
    valid_vectorlist,
)

logger = logging.getLogger(__name__)


@valid_input(dict={"rst_in": valid_raster})
def check_rst_dimensions(rst_in, minmax, ncols, nrows, transform=None):
    """This function checks if the input raster has the desired dimensions.

    Parameters
    ----------
    rst_in: str
        File path to input raster
    minmax: list
        Containing xmin, ymin, xmax, ymax
    ncols: int
        Number of columns in the raster
    nrows: int
        Number of rows in the raster
    transform:  rasterio.transform, default None
        Transformation as defined in Rasterio.

    Returns
    -------
    bool
        Raster has the required dimensions (True/False)
    """
    coordinates, transf, cols, rows = read_rst_params(rst_in)

    if cols != ncols:
        return False

    elif rows != nrows:
        return False

    elif minmax != coordinates:
        return False

    elif transform is not None:
        if transf != transform:
            return False
    else:
        return True


@valid_input(dict={"rst_in": valid_raster})
def read_rst_params(rst_in):
    """Read all spatial dimensions of a raster

    Parameters
    ----------
    rst_in: pathlib.Path
        File path to the input raster

    Returns
    -------
    minmax: list
        A list with the extreme coordinate values (xmin, ymin, xmax and ymax)
    transform: rasterio.transform
        Transformation as defined in Rasterio.
    cols: int
        The number of columns in the raster
    rows: int
        The number of rows in the raster
    """
    with rasterio.open(rst_in) as src:
        cols = src.width
        rows = src.height
        xmin = src.bounds[0]
        ymin = src.bounds[1]
        xmax = src.bounds[2]
        ymax = src.bounds[3]
        transf = src.transform
        minmax = [xmin, ymin, xmax, ymax]
    return minmax, transf, cols, rows


@valid_input(dict={"rst_in": valid_raster})
def read_dtype_raster(rst_in):
    """Read dtype of raster

    Parameters
    ----------
    rst_in: pathlib.Path
        File path to the input raster

    Returns
    -------
    dtype: numpy.dtype

    Note
    -----
    Only works for single band rasters
    """
    with rasterio.open(rst_in) as src:
        dtype = src.dtypes[0]
    return np.dtype(dtype)


@valid_input(dict={"rst_in": valid_raster})
def read_rasterio_profile(rst_in):
    """Read all spatial dimensions of a raster

    Parameters
    ----------
    rst_in: pathlib.Path
        File path to the input raster

    Returns
    -------
    profile: dict
        See :func:`pywatemsedem.geo.rasterproperties.RasterProperties.rasterio_profile`
    """
    with rasterio.open(rst_in) as src:
        profile = src.profile

    return profile


def write_arr_as_rst(arr, rst_out, dtype, profile):
    """Write numpy.ndarray as a raster-file

    Parameters
    ----------
    arr: numpy.ndarray
        2D numpy array to be written as a raster file
    rst_out: str
        File path to the output raster
    dtype: numpy.dtype
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`
    """
    mandatorykeys = [
        "driver",
        "height",
        "width",
        "crs",
        "transform",
        "count",
        "nodata",
    ]
    for key in mandatorykeys:
        if key not in profile.keys():
            msg = "not all mandatory keys in the profile are given! "
            msg += key
            raise Exception(msg)
    if "compress" not in profile:
        profile["compress"] = "DEFLATE"

    if "dtype" in profile.keys():
        profile.pop("dtype", None)

    if not rasterio.dtypes.check_dtype(dtype):
        msg = f"no valid dtype chosen! {dtype}"
        raise TypeError(msg)

    if arr.shape[1] != profile["width"] or arr.shape[0] != profile["height"]:
        msg = "dimensions of array are not the same of the given profile"
        raise Exception(msg)

    with rasterio.open(rst_out, "w", dtype=dtype, **profile) as dst:
        dst.write(arr, 1)


def check_raster_properties_raster_with_template(path_check, path_template, epsg):
    """Checks if extent and resolution of new raster and template raster align

    Parameters
    ----------
    path_check: pathlib.Path | RasterProperties
                Incoming, new raster
    path_template: pathlib.Path
                    Template raster
    epsg: int
            EPSG code of the cartographic projection
    """
    if isinstance(path_check, Path):
        rp_check = RasterProperties.from_rasterio(
            read_rasterio_profile(path_check), epsg=epsg
        )
    else:
        rp_check = path_check

    rp_template = RasterProperties.from_rasterio(
        read_rasterio_profile(path_template), epsg=epsg
    )
    if rp_check.bounds != rp_template.bounds:
        msg = "Extent of new raster not the same as of template raster"
        raise ValueError(msg)
    elif rp_check.resolution != rp_template.resolution:
        msg = "Resolution of new raster not the same as of template raster"
        raise ValueError(msg)


@valid_input(dict={"rst_in1": valid_raster, "rst_in2": valid_raster})
def grid_difference(rst_in1, rst_in2, rst_out):
    """Make the difference between two grids

    Parameters
    ----------
    rst_in1: str
        File path to inputraster 1
    rst_in2: str
        File path to inputraster 2
    rst_out: str
        File path to outputraster
    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "grid_calculus", "3"]
    cmd_args += ["-A", str(rst_in1), "-B", str(rst_in2), "-C", str(rst_out)]

    execute_saga(cmd_args)


@valid_input(dict={"rst_in": valid_raster})
def create_hillshade(rst_in, rst_hillshade):
    """Create a hillshade raster of the DTM

    Parameters
    ----------
    rst_in: str
        File path of input raster
    rst_hillshade: str
        File path of output raster
    """
    if isinstance(rst_hillshade, str):
        rst_hillshade = Path(rst_hillshade).parent / (Path(rst_hillshade).stem + ".tif")
    elif isinstance(rst_hillshade, Path):
        rst_hillshade = rst_hillshade.parent / (rst_hillshade.stem + ".tif")
    else:
        msg = (
            f"'hillshade' is of type {type(rst_hillshade)}, "
            f"cannot be converted to Pathlib Path"
        )
        raise TypeError(msg)

    cmd_args = ["gdaldem", "hillshade", "-q"]
    cmd_args += ["-co", "COMPRESS=DEFLATE"]
    cmd_args += [str(rst_in), str(rst_hillshade)]
    execute_subprocess(cmd_args)


@valid_input(dict={"vct": valid_vector})
def get_fields_vct(vct):
    """Get a list of all fields in a shapefile

    Parameters
    ----------
    vct: str
        File path to shapefile

    Returns
    -------
    list
        Field names of the shape/vector-file.
    """
    with fiona.open(vct) as c:
        fields = c.schema["properties"].keys()
        return fields


@valid_input(dict={"vct": valid_vector})
def get_geometry_type(vct):
    """Get the geometry type of a shapefile

    Parameters
    ----------
    vct: str
        File path to shapefile

    Returns
    -------
    str
        Geometry type (see fiona documentation for all possibilities)
    """
    with fiona.open(vct) as c:
        geom = c.schema["geometry"]
    return geom


def copy_rst(rst_in, rt_out):
    """Copy a raster and converts it to an idrisi-raster

    Parameters
    ----------
    rst_in: str
        File path of input raster
    rt_out: str
        File path of output raster (extension must be .rst!)

    Note
    -----
    Uses and relies on gdal_translate CLI
    """
    if rt_out.exists():
        delete_rst(rt_out)

    cmd_args = ["gdal_translate", "-q", "-of", "RST", str(rst_in), str(rt_out)]
    execute_subprocess(cmd_args)


@valid_input(dict={"vct_in": valid_vector})
def copy_vct(vct_in, folder_out):
    """Copies a shapefile to another location

    Parameters
    ----------
    vct_in: str or pathlib.Path
        File path of the shapefile to be copied.
    folder_out: str or pathlib.Path
        File path of the destination shapefile.

    Note
    -----
    Uses and relies on ogr2ogr CLI
    """
    cmd_args = ["ogr2ogr", str(folder_out), str(vct_in)]
    execute_subprocess(cmd_args)


@valid_input(dict={"tiff_in": valid_raster})
def tiff_to_esri_shp(tiff_in, vct_out, epsg):
    """
    Transform a tiff file to an esri shape (raster to shape)

    Parameters
    ----------
    tiff_in: str or pathlib.Path
        name input tiff
    vct_out: str or pathlib.Path
        name input shapefile
    epsg: str, default None
        the epsg code defining the coordinate system of the raster,
        format = "EPSG:XXXXX"
    """
    gpd_db = tiff_to_geopandas_df(tiff_in, epsg)
    gpd_db.to_file(Path(vct_out))


@valid_input(dict={"tiff_in": valid_raster})
def tiff_to_geopandas_df(tiff_in):
    """Transform a tiff file to a geopandas dataframe

    Parameters
    ----------
    tiff_in: str or pathlib.Path
        File path of tiff raster file

    Returns
    -------
    gdf: geopandas.GeoDataFrame
        Geopandas representation of tiff raster
    """
    with rasterio.open(str(tiff_in)) as src:
        image = src.read(1)
        crs = src.profile["profile"]
        results = (
            {"properties": {"values": values}, "geometry": geometry}
            for i, (geometry, values) in enumerate(
                shapes(image, transform=src.profile["transform"])
            )
        )

    geoms = list(results)
    gdf = gpd.GeoDataFrame.from_features(geoms, crs=crs)

    return gdf


@valid_input(dict={"rst_in": valid_raster})
def idrisi_to_tiff(rst_in, tiff_out, dtype, epsg="EPSG:31370"):
    """Convert an Idrisi RST to GeoTiff

    Parameters
    ----------
    rst_in: str
        File path of the input rst.
    tiff_out: str
        File path of the output tiff file.
    dtype: str
        Data type of the destination rst, either 'integer' or 'float'.

    Note
    -----
    Uses and relies on gdal_translate CLI
    """

    if dtype == "integer":
        cmd_args = [
            "gdal_translate",
            "-q",
            "-ot",
            "Int16",
            "-of",
            "GTiff",
            "-co",
            "COMPRESS=DEFLATE",
            "-a_srs",
            str(epsg),
            str(rst_in),
            str(tiff_out),
        ]
        execute_subprocess(cmd_args)
    elif dtype == "float":
        cmd_args = [
            "gdal_translate",
            "-q",
            "-ot",
            "Float32",
            "-of",
            "GTiff",
            "-co",
            "COMPRESS=DEFLATE",
            "-a_srs",
            str(epsg),
            str(rst_in),
            str(tiff_out),
        ]
        execute_subprocess(cmd_args)
    else:
        msg = 'Saving rst as tif. Type must be "integer" or "float"'
        logger.error(msg)
        raise TypeError(msg)


def valid_gdal_type(func):
    """Check if your input array mask is valid. Use as decorator"""

    @wraps(func)
    def wrapper(*args, dtype):
        """Wrapper function

        Parameters
        ----------
        *args:
            non keyword arguments
        dtype: numpy.dtype or str
            allowed types are "Float32", "Float64", "Int16", "Int32", "Int64"
            integer is Int16, float is Float32
        """
        if dtype == "integer":
            dtype = "Int16"
        if dtype == "float":
            dtype = "Float32"
        implemented_types = ["Float32", "Float64", "Int16", "Int32", "Int64"]
        # check value dtype and if not string
        if dtype in [np.int16, np.int32, np.int64, np.float32, np.float64]:
            try:
                dtype = dtype.__name__
            except AttributeError:
                dtype = dtype.name
        dtype = dtype.capitalize()
        if dtype not in implemented_types:
            raise IOError(
                f"Raster type {dtype} not known for gdal, please select "
                f"{', '.join(implemented_types)}"
            )
        func(*args, dtype)

    return wrapper


@valid_input(dict={"tiff_in": valid_raster})
@valid_gdal_type
def tiff_to_idrisi(tiff_in, rst_out, dtype):
    """Convert a GeoTiff to an Idrisi RST

    Parameters
    ----------
    tiff_in: str
        File path of the input tiff file
    rst_out: str
        File path of the destination rst
    dtype: str, default Float64
        Raster type

    Note
    ----
    Uses and relies on gdal_translate CLI
    """
    cmd_args = [
        "gdal_translate",
        "-q",
        "-ot",
        dtype.capitalize(),
        "-of",
        "RST",
        str(tiff_in),
        str(rst_out),
    ]
    execute_subprocess(cmd_args)


def get_feature_count(vct):
    """Count the amount of features in a shapefile

    Parameters
    ----------
    vct: str
        File path of the shapefile.

    Returns
    -------
    nr_features: int
        Number of features in the shapefile
    """
    with fiona.open(vct) as c:
        nr_features = len(c)
    return nr_features


@valid_input(dict={"lst_vct": valid_vectorlist})
def merge_lst_vct(lst_vct, vct_out, epsg):
    """Merge a list of shapefiles to one shapefile

    Parameters
    ----------
    lst_vct: list
        List with file paths (str) of all shapefiles to be merged.
    vct_out: str
        File path of merged vct
    epsg: str
        The epsg code defining the coordinate system of the raster,
        format = "EPSG:XXXXX"

    Note
    ----
    Uses and relies on ogr2ogr CLI
    """
    for i, vct in enumerate(lst_vct):
        if i == 0:
            copy_vct(vct, vct_out)
        else:
            cmd_args = ["ogr2ogr", "-update", "-a_srs", str(epsg)]
            cmd_args += ["-append", str(vct_out)]
            cmd_args.append(str(vct))
            cmd_args += ["-nln", vct_out.stem]

            execute_subprocess(cmd_args)


@valid_input(dict={"lst_vct": valid_vectorlist})
def merge_lst_vct_saga(lst_vct, vct_out):
    """Merge a list of shapefiles to one shapefile

    Parameters
    ----------
    lst_vct: list
        List with file paths (str) of all shapefiles to be merged.
    vct_out: str
        File path + name of the destination shapefile.

    Note
    ----
    Uses and relies on ogr2ogr CLI
    """

    cmd_args = (
        ["saga_cmd", SAGA_FLAGS, "shapes_tools", "2"]
        + [f"-INPUT={';'.join(lst_vct)}"]
        + [f"-MERGED={vct_out}"]
    )
    execute_saga(cmd_args)


@valid_input(dict={"rst_in": valid_raster})
def delete_rst(rst_in):
    """Delete a raster dataset

    Parameters
    ----------
    rst_in: str
        File path of the raster dataset to be deleted

    """
    cmd_args = ["gdalmanage", "delete", str(rst_in)]
    execute_subprocess(cmd_args)
    return


@valid_input(dict={"rst_in": valid_raster})
def clip_rst(rst_in, rst_out, Cnst, resampling="near"):
    """Clips a raster to a certain bounding box with a given resolution

    Parameters
    ----------
    rst_in: str
        File path to in input raster.
    rst_out: str
        File path to the destination raster.
    Cnst: dict
        Dictionary with following keys:

        - *epsg* (str): the EPSG-code of the rst_in
        - *res* (int): resolution
        - *nodata* (int): nodata flag
        - *minmax* (list): list with xmin, ymin, xmax, ymax
    resampling: str, default 'near'
        Either "mode" or "near"

    Notes
    -----
    1. "mode" and "near" have been tested, see
        https://gdal.org/programs/gdalwarp.html#cmdoption-gdalwarp-r
    """

    rst_in = Path(rst_in)
    rst_out = Path(rst_out)

    logger.info(f"Clipping {rst_in.name}...")
    cmd_args = ["gdalwarp", "-q", "-s_srs", str(Cnst["epsg"])]
    cmd_args += ["-t_srs", str(Cnst["epsg"])]
    cmd_args += ["-te"]
    for oor in Cnst["minmax"]:
        cmd_args += [str(oor)]
    cmd_args += ["-tr", str(Cnst["res"]), str(Cnst["res"])]
    cmd_args += ["-r", resampling]
    cmd_args += [str(rst_in), str(rst_out)]
    execute_subprocess(cmd_args)


@valid_input(dict={"vct": valid_vector})
def get_extent_vct(vct):
    """Gets the bounding box coordinates of a shapefile

    Parameters
    ----------
    vct: str or pathlib.Path
        File path of the input shapefile.

    Returns
    -------
    tuple
        xmin, ymin, xmax, ymax
    """
    with fiona.open(vct) as c:
        extent = c.bounds
    return extent  # xmin, ymin, xmax, ymax


@valid_input(dict={"vct": valid_vector, "vct_clip": valid_vector})
def clip_vct(vct_in, vct_out, vct_clip, overwrite=False, lst_ignore_field=None):
    """Clip a shapefile by another shapefile

    Parameters
    ----------
    vct_in: pathlib.Path
        File path of shapefile to be clipped.
    vct_out: pathlib.Path
        File path of the destination shapefile
    vct_clip: str
        File path of the clip boundary vector
    overwrite: bool, default False
        If True, overwrite existing file
    lst_ignore_field: list, optional
        Ignore a specific field from input in output.

    Note
    ----
    Uses and relies on ogr2ogr CLI
    """
    cond = True
    if not overwrite:
        if vct_out.exists():
            cond = False
    if cond:
        logger.info(f"Clipping {vct_in.name}...")
        ext = get_extent_vct(vct_clip)
        cmd_args = [
            "ogr2ogr",
            "-spat",
            str(ext[0]),
            str(ext[1]),
            str(ext[2]),
            str(ext[3]),
        ]
        cmd_args += ["-skipfailures"]
        cmd_args += ["-clipsrc", str(vct_clip)]
        if lst_ignore_field is not None:
            fields = get_fields_vct(vct_in)
            cmd_args += [
                "-select",
                (", ").join(
                    [field for field in fields if field not in lst_ignore_field]
                ),
            ]
        cmd_args += [str(vct_out), str(vct_in)]
        execute_subprocess(cmd_args)


@valid_input(dict={"lst_rasters": valid_rasterlist, "vct_polygon": valid_vector})
def compute_statistics_rasters_per_polygon_vector(
    lst_rasters,
    vct_polygon,
    vct_out,
    lst_names,
    dict_operators,
    normalize=True,
    ton=False,
):
    """Compute statistics

    Parameters
    ----------
    lst_rasters: list
        File path rasters (pathlib.Path)
    vct_polygon: str
        File path to input polygon vector.
    vct_out: str
        File path to output polygon vector.
    lst_names: dict
        List of output names in output vector for files in lst_rasters.
    dict_operators: dict
        Operators. For a description of the dictionary of statistics, see
        inputs in :func:`pywatemsedem.geo.utils.grid_statistics`. See example for use.
    normalize: bool, default False
        Normalize with shape area (True)
    ton: bool, default False
        Use ton (true)

    Note
    ----
    The desired statistics are inputted after the file reference of the raster, and are
    formatted as a dictionary, e.g.:

    .. code-block:: python

        dict_operators = {"COUNT":True,"SUM":True}

    Examples
    --------

    >>> vct_aho = "AHO.shp"
    >>> vct_out = "statistics_aho.shp"
    >>> rst_sewerin ="sewerin.rst"
    >>> rst_sediexport ="SediExport.rst"
    >>> compute_statistics_rasters_per_polygon_vector(vct_aho,
    >>>                                                   vct_out,
    >>>                                                   [rst_sewerin,rst_sediexport]
    >>>                                                   ["River","Sewers"],
    >>>                                                   {"COUNT":True,"SUM":True},
    >>>                                                   ton = True)
    """
    grid_statistics(lst_rasters, vct_polygon, vct_out, **dict_operators)

    gdf = gpd.read_file(vct_out)
    rename = {
        f"G{ind + 1:02d}_{op}": name
        for ind, name in enumerate(lst_names)
        for op in dict_operators
    }
    gdf = gdf.rename(columns=rename)
    for col in lst_names:
        gdf.loc[gdf[col].isnull(), col] = 0
        if ton:
            gdf[col] = gdf[col] / 1000.0
        if normalize:
            gdf[col + "_ha"] = gdf[col] * 100**2 / (gdf.area)

    gdf.to_file(vct_out)

    return gdf


@valid_input(dict={"rst_in": valid_raster})
def rst_to_vct_points(rst_in, vct_out):
    """Convert all no nodata values in a raster to a vector point file.

    Parameters
    ----------
    rst_in: str or pathlib.Path
        File path of the raster file that should be converted to a shapefile.
    vct_out: pathlib.Path
        File path of the destination vct.
    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "shapes_grid", "3"]
    cmd_args += ["-GRIDS", str(rst_in)]
    cmd_args += ["-SHAPES", str(vct_out)]
    execute_saga(cmd_args)


@valid_input(dict={"vct_in": valid_vector})
def vct_to_rst_value(
    vct_in, rst_out, rstval, Cnst, nodata=-9999, alltouched=True, dtype=None
):
    """Rasterizes a shapefile by a given constant value

    Parameters
    ----------
    vct_in: str or pathlib.Path
        File path of the shapefile to be rasterized.
    rst_out: pathlib.Path
        File path of the destination rst
    Cnst: dict
        Dictionary with following keys:

        - *res* (int): resolution
        - *nodata* (int): nodata flag
        - *minmax* (list): list with xmin, ymin, xmax, ymax

    rstval: str
        The value all features of the shapefile will get in the raster.
    alltouched: bool, default true
        Enables the ALL_TOUCHED rasterization option so that all pixels
        touched by lines or polygons will be updated.
    dtype: str, default None
        Data type of the values, e.g. Byte/Int16/UInt16/UInt32/Int32/Float32...

    Note
    -----
    Uses and relies on gdal_rasterize CLI
    """
    cmd_args = ["gdal_rasterize", "-q", "-a_nodata", str(nodata)]
    if alltouched:
        cmd_args += ["-at"]
    cmd_args += ["-burn", str(rstval)]
    cmd_args += ["-l", vct_in.stem]
    cmd_args += ["-of", "GTiff", "-te"]
    for oor in Cnst["minmax"]:
        cmd_args += [str(oor)]
    if dtype is not None:
        cmd_args += ["-ot", dtype]
    cmd_args += ["-tr", str(Cnst["res"]), str(Cnst["res"])]
    cmd_args += ["-co", "COMPRESS=DEFLATE"]
    cmd_args += [str(vct_in), str(rst_out)]

    execute_subprocess(cmd_args)


@valid_input(dict={"rst_in": valid_raster})
def reclass_rst(rst_in, rst_out, df_reclass):
    """Reclass of a raster based on pandas dataframe

    Parameters
    ----------
    rst_in: str or pathlib.Path
        File path of the input raster
    rst_out: pathlib.Path
        File path of the destination rst (.sdat)
    df_reclass: pandas.DataFrame
        DataFrame containing the mapping with the columns:

        - *RST_VAL*: current values
        - *NEWVAL*: new value

    Note
    -----
    Uses and relies on saga_cmd CLI
    """
    df_reclass["MAX"] = df_reclass["RST_VAL"] + 1
    txt_reclass = rst_out.parent / (rst_out.stem + "_reclasstable.csv")
    df_reclass[["RST_VAL", "MAX", "NEWVAL"]].to_csv(txt_reclass, index=False)
    cmd_args = ["saga_cmd", SAGA_FLAGS, "grid_tools", "15", "-INPUT", str(rst_in)]
    cmd_args += ["-RESULT", str(rst_out), "-METHOD", "3", "-RETAB_2", str(txt_reclass)]
    cmd_args += [
        "-F_MIN=RST_VAL",
        "-F_MAX=MAX",
        "-F_CODE=NEWVAL",
        "-TOPERATOR=0",
        "-RESULT_NODATA_CHOICE=0",
    ]

    execute_saga(cmd_args)


@valid_input(dict={"rst_in": valid_raster})
def raster_to_polygon(rst_in, vct_out):
    """Polygonize a raster

    This function converts rastercells to polygons

    Parameters
    ----------
    rst_in: str or pathlib.Path
        File path of theinput raster
    vct_out: str or pathlib.Path
        File path of the destination shapefile

    Note
    -----
    Uses and relies on saga_cmd CLI
    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "shapes_grid", "6", "-GRID", str(rst_in)]
    cmd_args += ["-POLYGONS", str(vct_out), "-CLASS_ALL", "1", "-SPLIT", "0"]
    execute_saga(cmd_args)


@valid_input(dict={"vct_lines": valid_linesvector, "vct_points": valid_pointvector})
def lines_to_points(vct_lines, vct_points, distance):
    """Converts a shapefile with lines to a shapefile with points.

    Parameters
    ----------
    vct_lines: str or pathlib.Path
        File path to the input shapefile with lines.
    vct_points: str or pathlib.Path
        File path to the output shapefile with points.
    distance: float
        Distance between two points

    Note
    -----
    Uses and relies on saga_cmd CLI
    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "shapes_points", "5"]
    cmd_args += ["-LINES", str(vct_lines), "-POINTS", str(vct_points)]
    cmd_args += ["-ADD", "1", "-METHOD_INSERT", "0", "-DIST", distance]
    cmd_args += ["-ADD_POINT_ORDER", "1"]

    execute_saga(cmd_args)
    return


@valid_input(dict={"vct_line": valid_linesvector, "rst_template": valid_raster})
def lines_to_direction(vct_line, rst_out, rst_template):
    """Converts line features to a direction raster

    This function converts the direction of line features to a raster. See the
    :ref:`docs of cnws for more information <watemsedem:routingmap>` about this raster

    Parameters
    ----------
    vct_line: str or pathlib.Path
        File path to the input line shapefile
    rst_out: pathlib.Path
        File path to the output raster
    rst_template: str or pathlib.Path
        File path to a template raster

    Note
    -----
    Uses and relies on saga_cmd CLI

    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "line_direction", "0"]
    cmd_args += [
        "-INPUT",
        str(vct_line),
        "-TARGET_DEFINITION",
        "1",
        "-TARGET_TEMPLATE",
        str(rst_template),
    ]
    cmd_args += ["-GRID", str(rst_out), "-ORDER_FIELD", "sort_order"]

    execute_saga(cmd_args)

    rst_out = rst_out.parent / (rst_out.stem + ".sdat")

    arr, profile = load_raster(rst_out)
    profile["compress"] = "DEFLATE"
    profile["nodata"] = 0
    arr += 1
    arr = np.where(arr == 256, 0, arr).astype("int16")
    write_arr_as_rst(arr, rst_out, "int16", profile)
    return


@valid_input(dict={"vct": valid_vector})
def check_single_polygon(vct):
    """Check if the catchment polygon is a single polygon (not empty or
    multipolygon).

    Parameters
    -----------
    vct: str or pathlib.Path
        File path of the shapefile
    """
    with fiona.open(vct) as c:
        nr_polygons = len(c)
        if nr_polygons != 1:
            msg = (
                f"Catchment polygon should be a single polygon, current "
                f"catchment polygon holds '{nr_polygons}' polygons."
            )
            IOError(msg)


@valid_input(dict={"vct_in": valid_vector})
def vct_to_rst_field(
    vct_in, rst_out, Cnst, field=None, alltouched=True, dtype="Float64"
):
    """Rasterizes a shapefile by a given attribute field

    Parameters
    ----------
    vct_in: str or pathlib.Path
        File path of the shapefile to be rasterized.
    rst_out: pathlib.Path
        File path of the destination rst
    Cnst: dict
        Dictionary with following keys:

        - *res* (int): resolution
        - *nodata* (int): nodata flag
        - *minmax* (list): list with xmin, ymin, xmax, ymax

    field: str
        The field of the shapefile containing the values for the raster
    alltouched: bool, default True
        Enables the ALL_TOUCHED rasterization option so that all pixels
        touched by lines or polygons will be updated.
    dtype: str, default None
        Data type of the values, e.g. Byte/Int16/UInt16/UInt32/Int32/Float32...

    Note
    ----
    Uses and relies on gdal_rasterize CLI
    """
    # if rst_out.exists():
    #    delete_rst(rst_out)

    if isinstance(rst_out, str):
        rst_out = Path(rst_out).parent / (Path(rst_out).stem + ".tif")
    elif isinstance(rst_out, Path):
        rst_out = rst_out.parent / (rst_out.stem + ".tif")
    else:
        msg = (
            f"'outrst' is of type {type(rst_out)}, "
            f"can not be converted to Pathlib Path"
        )
        raise TypeError(msg)

    if isinstance(vct_in, str):
        vct_in = Path(vct_in).parent / (Path(vct_in).stem + ".shp")
    elif isinstance(vct_in, Path):
        vct_in = vct_in.parent / (vct_in.stem + ".shp")
    else:
        msg = (
            f"'inshp' is of type {type(vct_in)}, "
            f"can not be converted to Pathlib Path"
        )
        raise TypeError(msg)

    cmd_args = ["gdal_rasterize", "-q", "-a_nodata", str(Cnst["nodata"])]
    if alltouched:
        cmd_args += ["-at"]
    if field:
        cmd_args += ["-a", field]
    cmd_args += ["-l", vct_in.stem]
    cmd_args += ["-of", "GTiff", "-te"]
    for oor in Cnst["minmax"]:
        cmd_args += [str(oor)]

    cmd_args += ["-ot", dtype]
    cmd_args += ["-tr", str(Cnst["res"]), str(Cnst["res"])]
    cmd_args += ["-co", "COMPRESS=DEFLATE"]
    cmd_args += [str(vct_in), str(rst_out)]

    execute_subprocess(cmd_args)


def execute_saga(cmd_args):
    """Run saga executable and catch non-informative error

    This function catches saga error command and ignores 'corrupted size vs. prev_size
    in fastbins' error in case output runs from saga are okay.

    Parameters
    ----------
    saga_cmd: list
        Saga command
    """
    if "saga_cmd" not in cmd_args:
        msg = f"{' '.join(cmd_args)} is not a saga command."
        raise IOError(msg)
    try:
        subprocess.check_output(cmd_args, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        if ("corrupted size vs. prev_size in fastbins" in e.output.decode()) & (
            "okay" in e.output.decode()
        ):
            pass
        else:
            print(e.output.decode())


@valid_input(dict={"vct_polygon": valid_polygonvector, "rst_template": valid_raster})
def polygons_to_raster(vct_polygon, rst_out, rst_template, field, dtype):
    """Converts a polygon shapefile to a raster

    Parameters
    ----------
    vct_polygon: str or pathlib.Path
        File path of input polygon shapefile
    rst_out: str or pathlib.Path
        File path of output rasterfile
    rst_template: str or pathlib.Path
        File path to a template raster
    field: str
        The field of the shapefile containing the values for the raster
    dtype: str, default None
        Data type of the values, e.g. Byte/Int16/UInt16/UInt32/Int32/Float32...

    Note
    -----
    Uses and relies on saga_cmd CLI

    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "grid_gridding", "0"]
    cmd_args += ["-INPUT", str(vct_polygon), "-FIELD", str(field)]
    cmd_args += ["-OUTPUT", "2", "-POLY_TYPE", "1"]
    if dtype == "integer":
        grid_type = "6"  # 4 byte unsigned integer
    elif dtype == "float":
        grid_type = "7"  # 4 byte floating point
    else:
        grid_type = "9"  # same as attribute
    cmd_args += ["-GRID_TYPE", grid_type, "-TARGET_DEFINITION", "1"]
    cmd_args += ["-TARGET_TEMPLATE", str(rst_template), "-GRID", str(rst_out)]
    execute_saga(cmd_args)


@valid_input(dict={"vct_line": valid_linesvector, "rst_template": valid_raster})
def lines_to_raster(vct_line, rst_out, rst_template, field, dtype):
    """Converts a line shapefile to a raster

    Parameters
    ----------
    vct_line: str or pathlib.Path
        File path of input line shapefile
    rst_out: str or pathlib.Path
        File path of output rasterfile
    rst_template: str or pathlib.Path
        File path to a template raster
    field: str
        The field of the shapefile containing the values for the raster
    dtype: str, default None
        Data type of the values, e.g. Byte/Int16/UInt16/UInt32/Int32/Float32...

    Note
    -----
    Uses and relies on saga_cmd CLI

    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "grid_gridding", "0"]
    cmd_args += ["-INPUT", str(vct_line), "-FIELD", str(field)]
    cmd_args += ["-OUTPUT", "2", "-LINE_TYPE", "1"]
    if dtype == "integer":
        grid_type = "6"  # 4 byte unsigned integer
    elif dtype == "float":
        grid_type = "7"  # 4 byte floating point
    else:
        grid_type = "9"  # same as attribute
    cmd_args += ["-GRID_TYPE", grid_type, "-TARGET_DEFINITION", "1"]
    cmd_args += ["-TARGET_TEMPLATE", str(rst_template), "-GRID", str(rst_out)]

    try:
        execute_subprocess(cmd_args)
    except OSError as e:
        if "corrupted size vs. prev_size in fastbins" in str(e):
            try:

                rasterio.open(rst_out.with_suffix(".sdat"))
            except rasterio.errors.RasterioIOError:
                raise IOError(e)


@valid_input(dict={"vct_in": valid_vector})
def create_spatial_index(vct_in):
    """Creates a qix-file for a given shapefile

    Parameters
    ----------
    vct_in: str or pathlib.Path
        File path of the input shapefile

    Note
    -----
    Uses and relies on ogrinfo CLI
    """
    if vct_in.exists():
        cmd_args = ["ogrinfo", str(vct_in)]
        cmd_args += ["-q", "-sql", f"CREATE SPATIAL INDEX ON {vct_in.stem}"]
        execute_subprocess(cmd_args)
    else:
        msg = f"Shapefile does not exist! ({vct_in})"
        logger.error(msg)
        raise FileNotFoundError(msg)
    return


@valid_input(dict={"rst": valid_raster})
def load_raster(rst, return_bounds=False):
    """load raster with rasterio

    Parameters
    ----------
    rst: str or pathlib.Path
        File path of the file, .rst arr.
    return_bounds: bool, default False
        Flag to indicate whether a bounds of the arr should be returned.

    Returns
    -------
    arr: numpy.ndarray
        Array format of raster file.
    bounds: list
        List of bounds (xmin,ymin,xmax,ymax).
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`
    """
    # load
    try:
        with rasterio.open(rst) as src:
            arr = src.read()[0]
            profile = src.profile
            if return_bounds:
                bounds = src.bounds
    except rasterio.errors.RasterioIOError as e:
        logger.error(e)
        msg = f"could not open {rst}"
        raise Exception(msg)
    if return_bounds:
        return arr, profile, bounds
    else:
        return arr, profile


def raster_array_to_pandas_dataframe(arr_raster, profile):
    """Convert a raster array to a pandas dataframe.

    Parameters
    ----------
    arr_raster: numpy.ndarray
        Array raster format
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`

    Returns
    -------
    df: pandas.DataFrame
        A pandas format of the array raster with

        - *row* (int): the row id
        - *col* (int): the column id
        - *val* (float): the value
    """
    nrows = profile["height"]
    ncols = profile["width"]
    # convert to list
    arr = np.empty([nrows * ncols, 3], dtype=np.float32)
    arr[:, 0] = arr_raster.flatten()
    arr[:, 2] = np.tile(np.arange(1, ncols + 1, 1), nrows)
    arr[:, 1] = np.repeat(np.arange(1, nrows + 1, 1), ncols)
    df = pd.DataFrame(arr, columns=["val", "row", "col"])

    return df


def raster_dataframe_to_arr(df, profile, col, dtype):
    """Convert a pandas dataframe column to an array

    Parameters
    ----------
    df: pandas.DataFrame
        A pandas format of the array raster with

        - *row* (int): the row id
        - *col* (int): the column id
        - *val* (float): the value

    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`
    col: str
        Column name to convert to raster
    dtype: numpy.dtype

    Returns
    -------
    arr: numpy.ndarray
        Array of dataframe columns 'val'
    """
    nrows = profile["height"]
    ncols = profile["width"]
    if col in df:
        df[col] = deepcopy(df[col]).astype(dtype)
    else:
        raise ValueError(f"{col} not in list of raster")
    df = df.sort_values(["row", "col"])
    arr = np.reshape(df[col].values, (nrows, ncols))
    return arr


@valid_input(dict={"vct_a": valid_vector, "vct_b": valid_vector})
def saga_intersection(vct_a, vct_b, vct_intersect):
    """Calculate the intersection between two shapefiles

    Parameters
    ----------
    vct_a: str or pathlib.Path
        File path of input shapefile 1
    vct_b: str or pathlib.Path
        File path of input shapefile 2
    vct_intersect: str or pathlib.Path
        File path of the calculated intersection shapefile

    Note
    -----
    Uses and relies on saga_cmd CLI

    """

    cmd_args = ["saga_cmd", SAGA_FLAGS, "shapes_polygons", "14"]
    cmd_args += ["-A", str(vct_a)]
    cmd_args += ["-B", str(vct_b)]
    cmd_args += ["-RESULT", str(vct_intersect)]

    execute_saga(cmd_args)


@valid_input(dict={"vct": valid_vector})
def write_area_ha_to_vct(vct):
    """Add a field 'AREA_HA' to a shapefile

    This function calculates the area of every feature in a shapefile in hectare

    Parameters
    ----------
    vct: str or pathlib.Path
        File path to the input shapefile

    """
    gdf = gpd.read_file(vct)
    gdf["AREA_HA"] = gdf.area / 100.0**2
    gdf.to_file(vct)


@valid_input(dict={"vct_lines": valid_linesvector, "vct_polygons": valid_polygonvector})
def add_length_lines_to_polygons(vct_lines, vct_polygons, vct_out, name_field):
    """Calculate the total length of line segments within a polygon

    Parameters
    ----------
    vct_lines: str or pathlib.Path
        File path of the input line shapefile
    vct_polygons: str or pathlib.Path
        File path of the input polygon shapefile
    vct_out: str or pathlib.Path
        File path of the output polygon shapefile
    name_field: str
        Attribute name containing the lenght of the lines within a polygon

    """
    gdf_lines = gpd.read_file(vct_lines)
    gdf_poly = gpd.read_file(vct_polygons)
    gdf_lines = gdf_lines[["geometry"]]
    gdf_lines[name_field] = gdf_lines.length
    sjoin = gpd.sjoin(gdf_lines, gdf_poly[["geometry"]], how="inner", op="within")
    sjoin = sjoin.groupby(["index_right"]).sum()
    gdf_poly = pd.merge(gdf_poly, sjoin, how="left", left_index=True, right_index=True)
    gdf_poly.to_file(vct_out)


@valid_input(dict={"lst_rst": valid_rasterlist, "vct_in": valid_polygonvector})
def grid_statistics(
    lst_rst,
    vct_in,
    vct_out,
    naming=0,
    COUNT=False,
    MIN=False,
    MAX=False,
    RANGES=False,
    SUM=True,
    MEAN=False,
):
    """Calculate zonal statistics for a raster based on polygons

    Parameters
    ----------
    lst_rst: list
        File paths (str or pathlib.Path) of the input rasters
    vct_in: str or pathlib.Path
        File path of the input polygon shapefile
    vct_out: str or pathlib.Path
        File path of the output shapefile with the raster statistics
    naming: int, default 0
        1: grid name (note: esrsi shapes ar capped on 10 characters)
        0: grid id
    COUNT: bool, default False
        Count the raster cells within every polygon
    MIN: bool, default False
        Calculate the minimum value within every polygon
    MAX: bool, default False
        Calculate the maximum value within every polygon
    RANGES: bool, default False
        Calculate the range within every polygon
    SUM: bool, default False
        Calculate the sum of all pixel values within every polygon
    MEAN: bool, default False
        Calculate the mean of all pixel values within every polygon

    Note
    -----
    Uses and relies on saga_cmd CLI

    """
    cmd_args = ["saga_cmd", SAGA_FLAGS, "shapes_grid", "2"]
    cmd_args += ["-GRIDS", ";".join([str(i) for i in lst_rst])]
    cmd_args += ["-NAMING", str(naming)]
    cmd_args += ["-POLYGONS", str(vct_in)]
    cmd_args += ["-RESULT", str(vct_out)]
    a = ["-COUNT", "-MIN", "-MAX", "-RANGE", "-SUM", "-MEAN"]
    for j, opt in enumerate([COUNT, MIN, MAX, RANGES, SUM, MEAN]):
        cmd_args += [a[j]]
        if opt:
            cmd_args += ["1"]
        else:
            cmd_args += ["0"]

    cmd_args += ["-VAR", "0", "-STDDEV", "0"]
    execute_saga(cmd_args)


@valid_input(dict={"rst_in": valid_raster, "rst_mask": valid_raster})
def mask_raster(rst_in, rst_mask, un_id, folder="masked"):
    """Mask a raster by setting no data with no data positions of mask

    The input raster is masked by the nodata value in the mask raster.

    Parameters
    ----------
    rst_in: str or pathlib.Path
        File path raster to be masked
    rst_mask: str  or pathlib.Path
        File path mask raster, see note for format
    un_id: int
        unique id of the raster
    folder: str, default 'masked'
        Folder in which to safe masked raster

    Returns
    -------
    rst_masker: str
        File path of masked raster

    Note
    ----
    1. This mask raster should be a (**no** `nodata`, `nodata`)-raster in which the
       nodata  values indicate masking, and non-nodata values indicates not masking.
    2. Masking of the input raster is facilitated by setting to-mask-values in the
       input raster to `nodata`.

    """
    if folder not in os.listdir():
        os.mkdir(folder)
    arr_mask, profile_mask = load_raster(rst_mask, list_format=True)
    arr_in, profile_in = load_raster(rst_in, list_format=True)
    rst_masked = Path(rst_in)
    rst_masked = Path("temp") / f"{rst_masked.stem}-{un_id}{rst_masked.suffix}"

    arr_in[arr_mask == profile_mask["nodata"]] = profile_in["nodata"]
    write_arr_as_rst(arr_in, rst_masked, arr_in.dtype, profile_in)

    return rst_masked


@valid_input(dict={"lst_rst": valid_rasterlist})
def merge_rasters(lst_rst, rst_out, lst_rst_masks=None):
    """Merge several rasters to one raster with option to mask input rasters

    Parameters
    ----------
    lst_rst: list of str or list of pathlib.Path
        List of file paths of rasters which have to be merged together
    rst_out: str or pathlib.Path
        File path of merged raster
    lst_rst_masks: list
        List of file paths of mask files to be used to mask lst_rst_in, see
        :func:`pywatemsedem.geo.utils.mask_raster`.
    """
    if lst_rst_masks is not None:
        lst_rst_temp = []
        for i, rst_in in enumerate(lst_rst):
            lst_rst_temp.append(mask_raster(rst_in, lst_rst_masks[i], i, "temp"))
        lst_rst = lst_rst_temp

    lst_src = []
    for rst in lst_rst:
        src = rasterio.open(rst)
        lst_src.append(src)

    if len(lst_src) > 0:
        mosaic, out_trans = merge(lst_src)
        out_meta = src.meta.copy()
        out_meta.update(
            {
                "driver": "GTiff",
                "height": mosaic.shape[1],
                "width": mosaic.shape[2],
                "transform": out_trans,
                "compress": "DEFLATE",
            }
        )
        with rasterio.open(rst_out, "w", **out_meta) as dest:
            dest.write(mosaic)


def load_discharge_file(filename):
    """Function to read the discharge file as a dictionary

    Parameters
    ----------
    filename: str or pathlib.Path
        discharge file

    Returns
    -------
    discharge: dict
        contains discharge information
    """
    with open(filename) as f:
        lines = f.readlines()
    discharge = {}
    for line in lines[0:5]:
        discharge[line.split(":")[0]] = float(line.split(": ")[1].split(" (")[0])

    return discharge


def rasterprofile_to_rstparams(profile):
    """Transform rasterprofile to rstparams

    Parameters
    ----------
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`

    Returns
    -------
    rstparams: dict
        gdal dictionary holding all metadata for idrisi rasters

    """
    rstparams = {}

    rstparams["driver"] = "GTiff"
    minmax = profile["minmax"]
    rstparams["height"] = int(round((minmax[3] - minmax[1]) / profile["res"]))
    rstparams["width"] = int(round((minmax[2] - minmax[0]) / profile["res"]))
    rstparams["crs"] = profile["epsg"]
    rstparams["transform"] = rasterio.Affine(
        profile["res"],
        0,
        minmax[0],
        0,
        -profile["res"],
        minmax[3],
    )
    rstparams["count"] = 1
    rstparams["nodata"] = profile["nodata"]
    rstparams["compress"] = "DEFLATE"

    return rstparams


def rstparams_to_rasterprofile(rstparams, epsg=None):
    """Transform rstparams dictionary to to rasterio rasterprofile dictionary

    Parameters
    ----------
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`
    epsg: str, default None
        The epsg code defining the coordinate system of the raster,
        format = "EPSG:XXXXX"

    Returns
    -------
    rstparams: dict
        gdal dictionary holding all metadata for idrisi rasters

    """
    profile = {}

    profile["nodata"] = rstparams["nodata"]

    if "init" in list(rstparams["crs"].to_dict().keys()):
        profile["epsg"] = rstparams["crs"].to_dict()["init"]
    else:
        if epsg is not None:
            profile["epsg"] = epsg
        else:
            raise IOError("Cannot get epsg-code from rstparams, and no epsg defined!")

    profile["res"] = rstparams["transform"][0]

    xmin = rstparams["transform"][2]
    xmax = xmin + rstparams["transform"][0] * rstparams["width"]
    ymax = rstparams["transform"][5]
    ymin = ymax + rstparams["transform"][4] * rstparams["height"]

    profile["minmax"] = [xmin, ymin, xmax, ymax]
    profile["ncols"] = rstparams["width"]
    profile["nrows"] = rstparams["height"]

    return profile


@valid_input(dict={"rst_in": valid_raster})
def set_no_data_rst(
    rst_in, rst_out, arr_bindomain, profile, dtype=None, nodata_val=-9999
):
    """Set all pixels of a raster outside the model domain equal to NoData

    Parameters
    ----------
    rst_in: str
        File path of input raster to set no data values
    rst_out: str
        File path of output raster with no data values
    arr_bindomain: numpy.ndarray
        Array used to define domain nodata. In domain is equal to one,
        outside domain is equal to zero.
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`
    dtype: numpy.dtype, default None
        e.g. np.float64, np.float32, ..
    nodata_val: int, default -9999
        Standard value for nodata
    """

    profile["nodata"] = nodata_val

    arr_out, _ = load_raster(rst_in)
    arr_out[arr_bindomain == 0] = profile["nodata"]

    arr_out, profile = set_dtype_arr_rst(arr_out, profile, dtype=dtype)

    write_arr_as_rst(arr_out, rst_out, arr_out.dtype, profile)


def valid_mask(func):
    """Check if your input array mask is valid. Use as decorator"""

    @wraps(func)
    def wrapper(arr, arr_mask, nodata):
        """Wrapper function

        Parameters
        ----------
        arr:
            See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`
        arr_mask:
            See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`
        nodata:
            See :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`
        """
        # check values mask
        un_mask = np.unique(arr_mask)
        if set(un_mask) != set([1, 0]):
            msg = (
                f"Mask array should have values 1 (no mask) and 0 (mask)"
                f"(mask) values, instead of {un_mask}."
            )
            raise ValueError(msg)

        # check if size input array is equal to mask array
        if arr.size != arr_mask.size:
            msg = "Mask array has different size from input array"
            raise IOError(msg)

        # check if input array is not empty after masking
        cond = np.all(arr[arr_mask == 1] == nodata) or arr[arr_mask == 1].size == 0

        if cond:
            msg = (
                "Array after masking has only nodata values/is empty. Please check "
                "your mask/input array."
            )
            raise ValueError(msg)

        return func(arr, arr_mask, nodata)

    return wrapper


@valid_mask
def set_no_data_arr(arr, arr_mask, nodata):
    """Set no data to values to raster array outside mask.

    Parameters
    ----------
    arr: numpy.ndarray
        To mask array
    arr_mask: numpy.ndarray
        Mask array
    nodata: float
        The no data value to set in input array
    """
    arr = np.where(arr_mask == 1, arr, nodata)
    return arr


def set_dtype_arr_rst(arr, profile, dtype=None):
    """Set type for an array

    Parameters
    ----------
    arr: numpy.ndarray
        arr of a raster for which dtype has to be changed
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`
    dtype: numpy.dtype, default None
        e.g. np.float64, np.float32, ..

    Returns
    -------
    arr: numpy.ndarray
        dtype-updated arr of a raster for which dtype has to be changed
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile` with update dtype geo medata information
    """
    if dtype is not None:
        arr = arr.astype(dtype)
        profile["dtype"] = dtype

    return arr, profile


@valid_input(dict={"rst": valid_raster})
def calculate_sum_rst(rst):
    """Calculate sum of a raster values

    Parameters
    ----------
    rst: pathlib.Path
        File path of the raster file

    Returns
    -------
    float:
        sum of the values in the raster
    """
    arr, profile = load_raster(rst)

    return np.sum(arr[arr != profile["nodata"]])


def get_rstparams(CNWS_modelinputfolder, epsg=None, catchmentname="", template=None):
    """Get rstparams and rasterprofile from template raster (default:pkaart)

    Parameters
    ----------
    CNWS_modelinputfolder: str or pathlib.Path
        the path to the CNWS_modelinputfolder
    epsg: str, default None
        the epsg code defining the coordinate system of the raster,
        format = "EPSG:XXXXX"
    catchmentname: str, default ""
        catchment name
    template: str or pathlib.Path, default None
        File path to a template file that can be used as template for
        geodata and bin mask. Default the "P" raster is used.

    Returns
    -------
    profile: rasterio.profiles
        See :class:`rasterio.profiles.Profile`
    rstparams: dict
        gdal dictionary holding all metadata for idrisi rasters
    arr_bindomain: numpy.ndarray
        binary mask of modelling domain

    """
    # this template is used to generate binair mask
    if template is None:
        template = Path(CNWS_modelinputfolder) / "pfactor.rst"
    # open and assign profile to rstparams
    try:
        src = rasterio.open(template)
    except IOError:
        msg = (
            f"Templatefile '{template}' not known for getting spatial metadata "
            f"'rstparams' and 'profile'."
        )
        raise IOError(msg)
    rstparams = src.profile
    rstparams["compress"] = "DEFLATE"
    profile = rstparams_to_rasterprofile(rstparams, epsg=epsg)

    return rstparams, profile


def get_mask_template(CNWS_modelinputfolder, catchmentname, rst_template=None):
    """Get a binary raster from template raster (P-factor)

    Parameters
    ----------
    CNWS_modelinputfolder: str or pathlib.Path
        File path to the CNWS_modelinputfolder
    catchmentname: str
        Catchment name
    rst_template: str or pathlib.Path, default None
        Path to a template file that can be used as template for geodata and bin mask

    Returns
    -------
    arr_bindomain: numpy.ndarray
        In domain is equal to one, outside domain is equal to zero.
    """

    if rst_template is None:
        rst_template = Path(CNWS_modelinputfolder) / "pfactor.rst"
    try:
        src = rasterio.open(rst_template)
    except IOError:
        msg = (
            f"Templatefile '{rst_template}' not known for getting spatial metadata"
            f" 'rstparams' and 'rasterprofile'."
        )
        raise IOError(msg)
    # get bin_arr
    arr_bindomain = src.read()[0]
    arr_bindomain[arr_bindomain != src.profile["nodata"]] = 1
    arr_bindomain[arr_bindomain == src.profile["nodata"]] = 0

    return arr_bindomain


def any_equal_element_in_vector(geoseries_left, geoseries_right):
    """Check if there are equal vectors in the left and right geoseries

    Parameters
    ----------
    geoseries_left: geopandas.GeoSeries
    geoseries_right: geopandas.GeoSeries

    Return
    ------
    bool
        Line strings present in left and right geoseries
    """
    for x in geoseries_left:
        n = np.sum(geoseries_right.geom_equals(x))
        if n > 0:
            return True
    return False


@valid_input(dict={"rst_in": valid_raster})
def check_spatial_resolution_rst(rst_in, resolution, precision=0.01):
    """Check if the resolution of a tiff raster is equal to a defined one

    Parameters
    ----------
    rst_in: str or pathlib.Path
        File path of input raster
    resolution: int
        Resolution to compare with
    precision: float, default 0.01
        Precision to check resolution

    Returns
    -------
    cond: bool
        Equal/non-equal (True/False)
    """
    cond = True
    with rasterio.open(str(rst_in)) as src:
        res_tif = src.profile["transform"][0]
        if int(res_tif * 1 / precision) != int(resolution * 1 / precision):
            cond = False
    return cond


def estimate_width_of_polygon(
    arr_polygon_perimeter, arr_polygon_area, nan_value=np.nan
):
    """
    Estimate the width of a polygon with the length and area of the polygon
    see also
    https://gis.stackexchange.com/questions/20279/calculating-average-width-of-polygon/

    Parameters
    ----------
    arr_polygon_perimeter: numpy.ndarray or pandas.Series
        1D array holding in each row the perimeter of each polygon
    arr_polygon_area: numpy.ndarray or pandas.Series
        1D array holding in each row the area of each polygon
    nan_value: float
        The value to fill in for nan_values

    Returns
    -------
    arr_est_polygon_width: numpy.ndarray or pandas.Series
        1D array holding in each row the estimated width

    Note
    -----

    1. The width is only estimated for polygons which approximate a cuboid
    2. The width is estimated by

    .. math::

        B_gr =  1/4 [A-√(B-C)]
            1/4 [P_{poly}-√(〖P_{poly}〗^2-16A_{poly} )]

    with

    - :math:`P_{poly}` = perimeter of polygon
    - :math:`A_{poly}` = area of polygon

    See Also
    --------
    check_cuboid_condition

    """
    condition = check_cuboid_condition(arr_polygon_perimeter, arr_polygon_area)

    arr_est_polygon_width = np.ones(condition.shape) * nan_value

    arr_est_polygon_width[condition] = (
        arr_polygon_perimeter[condition]
        - np.sqrt(
            (arr_polygon_perimeter[condition] ** 2) - (16 * arr_polygon_area[condition])
        )
    ) / 4

    return arr_est_polygon_width


def check_cuboid_condition(arr_polygon_perimeter, arr_polygon_area):
    """Check if a shape can be classified as a cuboid

    Parameters
    ----------
    arr_polygon_perimeter: numpy.ndarray or pandas.Series
        1D array holding in each row the perimeter of each polygon
    arr_polygon_area: numpy.ndarray or pandas.Series
        1D array holding in each row the area of each polygon

    Returns
    -------
    condition: numpy.ndarray or pandas.Series
        1D array holding True/False if cuboid

    See Also
    --------
    estimate_width_of_polygon

    """
    condition = arr_polygon_perimeter**2 > 16 * arr_polygon_area

    return condition


def execute_subprocess(cmd_args):
    """Run a command line tool

    Logs error with command output and error

    Parameters
    ----------
    cmd_args: list
        list with all argmunts that must be given to the console

    Returns
    -------
    Returns True if function call is successfull.
    """
    try:
        logger.debug(cmd_args)
        subprocess.run(cmd_args, check=True, capture_output=True)
    except subprocess.CalledProcessError as e:
        logger.error(e.stdout.decode())
        logger.error(e.stderr.decode())
        raise OSError(
            f"Could not run '{cmd_args}', returning '{e.stderr.decode()}'-error"
        )

    return True


@valid_input(dict={"vct_in": valid_vector})
def clip_vct_with_bounds(vct_in, vct_out, bounds, overwrite=False):
    """Clip a shapefile using a bounding box


    Parameters
    ----------
    vct_in: pathlib.Path
        File path of shapefile to be clipped.
    vct_out: pathlib.Path
        File path of the destination shapefile
    bounds: list
        list with xmin, ymin, xmax, ymax
    overwrite: bool, default False
        if True, overwrite existing file

    Note
    -----
    Uses and relies on ogr2ogr CLI
    """
    cond = True
    if not overwrite:
        if vct_out.exists():
            cond = False
    if cond:
        logger.info("Clipping %s..." % vct_in.name)
        cmd_args = [
            "ogr2ogr",
            "-spat",
            str(bounds[0]),
            str(bounds[1]),
            str(bounds[2]),
            str(bounds[3]),
        ]
        cmd_args += ["-skipfailures"]
        cmd_args += [str(vct_out), str(vct_in)]
        execute_subprocess(cmd_args)


@valid_input(dict={"valid_catchment": valid_polygonvector})
def define_extent_from_vct(
    vct_catchment, resolution, nodata, epsg, bounds=None, buffer=100
):
    """Read the extent of the catchment

    Parameters
    ----------
    vct_catchment: str or pathlib.Path
        File path of the vector shapefile
    bounds: list, default None
        if None, bounds are determined from
        The **first** entry is **x_left**, the **second** entry is **y_lower**,
        the **third** entry is **x_right** and the **fourth** entry is **y_upper**
        [x_left, y_lower, x_right, y_upper].
    resolution: int
        Spatial resolution
    epsg: int
        EPSG code should be a numeric value, see https://epsg.io/.
    Returns
    -------
    rp: RasterProperties see :class:`pywatemsedem.geo.rasterproperties.RasterProperties`

    """
    with fiona.open(vct_catchment) as c:
        # Get spatial extent of catchment.
        extent = c.bounds
        minmax = [
            round(extent[0]) - buffer,
            round(extent[1]) - buffer,
            round(extent[2]) + buffer,
            round(extent[3]) + buffer,
        ]

    if bounds is not None:
        for i in range(4):
            if i < 2:  # xmin en ymin
                rest = (minmax[i] - bounds[i]) % resolution
                if rest != 0:
                    minmax[i] = minmax[i] - rest
            else:  # xmax en ymax
                rest = (minmax[i] - bounds[i - 2]) % resolution
                if rest != 0:
                    minmax[i] = minmax[i] - rest + resolution

    rp = RasterProperties(minmax, resolution, nodata, epsg)

    return rp


def clean_up_tempfiles(temporary_file, file_format):
    """Clean up extra generated tempfiles

    This function can be used to clean-up extra files (for example auxilary files)
    generated during a write of raster or vector file.

    Parameters
    ----------
    temporary_file: pathlib.Path | str

    file_format: {"tiff","shp","rst","txt"}
            format of the temporary files
    """
    if type(temporary_file) is str:
        temporary_file = Path(temporary_file)
    elif type(temporary_file) is tempfile._TemporaryFileWrapper:
        temporary_file.close()
        temporary_file = Path(temporary_file.name)
    if file_format == "tiff":
        check_suffix = SUFFIXES_TIF
    elif file_format == "rst":
        check_suffix = SUFFIXES_RST
    elif file_format == "shp":
        check_suffix = SUFFIXES_SHP
    elif file_format == "txt":
        check_suffix = SUFFIXES_TXT
    elif file_format == "saga":
        check_suffix = SUFFIXES_SAGA
    else:
        msg = f"File format '{file_format}' not implemented, cannot execute."
        raise NotImplementedError(msg)
    for suffix in check_suffix:
        filename = temporary_file.with_suffix(suffix)
        if filename.exists():
            filename.unlink()


def generate_vct_mask_from_raster_mask(rst_catchment, vct_catchment, resolution):
    """
    Generate a catchment fileshape from a raster file

    Parameters
    ----------
    rst_catchment: str or pathlib.Path
        Input raster format of the catchment (can be any type that can be read by
        rasterio)

    vct_catchment: str or pathlib.Path
        output shapefile format of the catchment

    resolution: int
        resolution of the model run
    """
    condition = check_spatial_resolution_rst(rst_catchment, resolution)
    if not condition:
        msg = (
            f"Defined resolution '{resolution}' is not the same as resolution "
            f"catchment raster '{rst_catchment}'. Aborting run."
        )
        raise IOError(msg)
    # gdf_catchment = tiff_to_geopandas_df(rst_catchment, crs)
    raster_to_polygon(rst_catchment, vct_catchment)
    gdf_catchment = gpd.read_file(vct_catchment)
    gdf_catchment = process_mask_shape_from_raster_file(gdf_catchment)
    gdf_catchment.to_file(Path(vct_catchment))


def process_mask_shape_from_raster_file(gdf_catchment, catchment_value=1):
    """
    Process the geopandas shape format of the input raster

    Parameters
    ----------
    gdf_catchment: geopandas.GeoDataFrame
        dataframe holding the mask , possible in multiple polygons (rows)
    catchment_value: int, default 1
        value that is used to define catchment

    Returns
    -------
    gdf_catchment: geopandas.GeoDataFrame
        dissolved dataframe holding mask, in one polygon (row)
    """
    gdf_catchment = gdf_catchment[gdf_catchment["VALUE"] == catchment_value]
    gdf_catchment = gdf_catchment.dissolve(by="VALUE")

    return gdf_catchment


def mask_array_with_val(arr, mask, mask_val):
    """Masking array by a mask

    Parameters
    ---------
    arr: numpy.ndarray
        array of which values are masked
    mask: numpy.ndarray
        masking array with same dimensions as arr
    mask_val: float or int
            masking arr when mask is mask_val
    """
    arr_masked = np.ma.masked_where(mask == mask_val, arr)
    return arr_masked


def nearly_identical(geoms, p, threshold=0.75):
    """Identify nearly identical geometries

    Parameters
    ----------
    geoms: geopandas.GeoSeries
    p: shapely.Geometry

    Returns
    -------
    pandas.Series
        True/False
    """

    nearly = (geoms.intersection(p).area / p.area) > threshold
    # return index values where nearly is True
    return pd.Series(nearly.index[nearly])



import inspect
import tempfile
from functools import wraps
from pathlib import Path

import fiona
import geopandas as gpd
import numpy as np
import rasterio
from fiona.collection import DriverError
from rasterio import RasterioIOError

from ..defaults import PREFIX_TEMP
from .rasterproperties import RasterProperties
from .rasters import RasterFile, RasterMemory, TemporalRaster
from .utils import (
    clean_up_tempfiles,
    define_extent_from_vct,
    generate_vct_mask_from_raster_mask,
    load_raster,
    vct_to_rst_value,
)
from .valid import PywatemsedemInputError, valid_exists
from .vectors import AbstractVector, VectorFile


def valid_mask_factory(func):
    """Check valid mask inputted when using raster or vectofactory"""

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        """Wrapper fun"""
        if self.mask is None:
            msg = (
                f"First create a mask with " f"{Factory.create_mask.__name__}-function"
            )
            raise PywatemsedemInputError(msg)
        return func(self, *args, **kwargs)

    return wrapper


class Factory:
    """Factory class

    Used to enable functions to generate vectors and rasters.

    Notes
    -----
    By default a rasterproperties instance is made in the initialisation
    See :func:`pywatemsedem.geo.factory.create_mask`-function. This can be toggled of by
    setting :const:`pywatemsedem.geo.factory.Factory.create_rasterproperties` to False
    """

    def __init__(self, resolution, epsg_code, nodata, resmap, bounds=None):
        """See class docs

        Parameters
        ---------
        resolution:int
            Model spatial resolution
        epsg_code: int
            EPSG code should be a numeric value, see https://epsg.io/.
        nodata: float
            See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
        resmap: pathlib.Path | str
            Folder path to write factory files to
        bounds: list, default None
            Raster boundaries which you wish for model.
            See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`
        """
        self._resolution = resolution
        self._epsg_code = epsg_code
        self._nodata = nodata
        self._bounds = bounds
        self._rp = None
        self._mask = None
        self._bounds = None
        resmap = Path(resmap) / "factory"
        if not resmap.exists():
            resmap.mkdir(exist_ok=True)
        self.mask_vector = resmap / "mask.shp"
        self.mask_raster = resmap / "mask.rst"
        # standard rasterproperties are generated in mask pahe from vector or raster
        # input
        self.create_rasterproperties = True

    @property
    def rp(self):
        "RasterProperties. See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`"
        return self._rp

    @rp.setter
    def rp(self, rasterproperties):
        "RasterProperties. See :class:`pywatemsedem.geo.rasterproperties.RasterProperties`"
        self._rp = rasterproperties

    @property
    def mask(self):
        "AbstractRaster mask"
        return self._mask

    @property
    def vct_mask(self):
        "AbstractVector mask, See :class:`pywatemsedem.geo.vectors.AbstractVector`"
        return self._vct_mask

    @mask.setter
    def mask(self, mask):
        """Set mask with an numpy array

        Parameters
        ----------
        mask : str or pathlib.Path
            Either a valid raster of vector file
        """
        self.create_mask(mask)

    def create_mask(self, mask):
        """Create mask based on a mask template (raster or vector)

        Parameters
        ----
        file_path: pathlib.Path | str
            File path to mask vector raster file

        Notes
        -----
        If :const:`pywatemsedem.geo.factory.Factory.create_rasterproperties` is set
        to False, one needs to self-define a RasterProperties instance.
        """
        valid_exists(mask, None)
        if self.create_rasterproperties is False:
            if self.rp is None:
                msg = (
                    f"Define a 'RasterProperties'-instance for the "
                    f"'{self.__class__.__name__}'-instance "
                    f"({self.__class__.__name__}.rp = RasterProperties(...)) or set "
                    f"'{self.__class__.__name__}.create_rasterproperties' to True."
                )
                raise IOError(msg)

        try:
            rasterio.open(mask)
        except RasterioIOError:
            try:
                fiona.open(mask)
            except DriverError:
                msg = "Input mask should be raster or vector polygon file."
                raise IOError(msg)
            else:
                if self.create_rasterproperties:
                    self.rp = define_extent_from_vct(
                        mask,
                        self._resolution,
                        self._nodata,
                        self._epsg_code,
                        self._bounds,
                    )
                tf_rst = tempfile.NamedTemporaryFile(
                    suffix=".tif", prefix=PREFIX_TEMP, delete=False
                )
                vct_to_rst_value(mask, tf_rst.name, 1, self.rp.gdal_profile)
                arr, _ = load_raster(tf_rst.name)
                clean_up_tempfiles(tf_rst, "tiff")
                self._vct_mask = VectorFile(mask)
        else:
            arr, rp = load_raster(mask)
            if self.create_rasterproperties:
                self.rp = RasterProperties.from_rasterio(rp, epsg=self._epsg_code)
            vct_mask = mask.with_suffix(".shp")
            generate_vct_mask_from_raster_mask(mask, vct_mask, self._resolution)
            self._vct_mask = VectorFile(vct_mask)
            self._vct_mask._geodata = self._vct_mask._geodata.set_crs(self.rp.epsg)

        self._mask = RasterMemory(arr, self.rp)
        self.vct_mask.write(self.mask_vector)
        self._mask.write(self.mask_raster, "idrisi")
        self._mask.arr_bin = np.where(
            self._mask.arr == self.rp.nodata, 0, self._mask.arr
        )

        return True

    @valid_mask_factory
    def raster_factory(
        self, raster_input, flag_clip=True, flag_mask=True, allow_nodata_array=False
    ):
        """Raster factory to load rasters in memory

        Parameters
        ----------
        raster_input: str, pathlib.Path or numpy.ndarray
            Input raster file or numpy array
        flag_clip: bool, default True
            Clip raster (True)
        flag_mask: bool, default True
            Mask raster (True)
        allow_nodata_array: default False
            Allow the returned array to only contian nodata-values,
            see :func:`pywatemsedem.geo.rasters.AbstractRaster.mask`.

        Returns
        -------
        raster: pywatemsedem.geo.rasters.AbstractRaster
            See :class:`pywatemsedem.geo.rasters.AbstractRaster`

        """
        arr_mask = self.mask.arr_bin if flag_mask else None
        if isinstance(raster_input, str):
            raster_input = Path(raster_input)
        if isinstance(raster_input, Path):
            try:
                rasterio.open(raster_input)
            except IOError:
                msg = (
                    f"Input raster file '{raster_input}' should be a valid raster "
                    f"file (e.g. IDRISI raster, geotiff, SAGA-GRID, ..)"
                )
                raise IOError(msg)
            rp = self.rp if flag_clip else None
            raster = RasterFile(
                raster_input, rp, arr_mask, allow_nodata_array=allow_nodata_array
            )
        elif isinstance(raster_input, np.ndarray):
            if raster_input.ndim == 2:
                raster = RasterMemory(
                    raster_input,
                    self.rp,
                    arr_mask,
                    allow_nodata_array=allow_nodata_array,
                )
            else:
                raster = TemporalRaster(raster_input, self.rp, arr_mask)
        else:
            print(type(raster_input))
            print(raster_input)
            m = inspect.currentframe()
            calframe = inspect.getouterframes(m, 2)
            [cal.function for cal in calframe]
            msg = (
                f"Input raster should be a numpy array or raster file, current type"
                f" is '{type(raster_input)}'"
            )
            raise IOError(msg)

        return raster

    @valid_mask_factory
    def vector_factory(self, vector_input, geometry_type, allow_empty=False):
        """Vector factory to load rasters in memory

        Parameters
        ----------
        vector_input: str, pathlib.Path or geopandas.GeoDataFrame
            Input vector file or geopandas dataframe
        mask: bool, default True
            Mask vector (True), nodata value will be that one of
            `pywatemsedem.geo.factory.Factory.rp`.
        allow_empty: bool, default False
            Allow vector to be empty, see :class:`pywatemsedem.geo.vectors.AbstractVector`

        Returns
        -------
        vector: pywatemsedem.geo.rasters.AbstractRaster
            See :class:`pywatemsedem.geo.rasters.AbstractRaster`
        """

        if isinstance(vector_input, str):
            vector_input = Path(vector_input)
        if isinstance(vector_input, Path):
            try:
                fiona.open(vector_input)
            except fiona.errors.DriverError:
                msg = (
                    f"Input vector file '{vector_input}' should be a valid "
                    f"vector file (e.g. ESRI shape file)."
                )
                raise IOError(msg)
            vector = VectorFile(
                vector_input, geometry_type, self.mask_vector, allow_empty=allow_empty
            )
        elif type(vector_input) == gpd.GeoDataFrame:
            vector = AbstractVector(
                vector_input, geometry_type, allow_empty=allow_empty
            )
        else:
            msg = "Input vector should be a geopandas GeoDataFrame or vector file."
            raise IOError(msg)

        return vector



from pathlib import Path

import pytest

from pywatemsedem.io.folders import check_and_create_folder


def test_check_and_create_folder(tmp_path):
    """Test creation of pywatemsedem folders"""
    foldername = Path(tmp_path) / "test"
    # check if it exists
    with pytest.raises(IOError) as excinfo:
        check_and_create_folder(foldername, f"Folder '{foldername}'")
    assert (f"Folder '{foldername}' does not exist.") in str(excinfo.value)

    # create it
    check_and_create_folder(foldername, f"Folder '{foldername.stem}'", create=True)
    assert foldername.exists()

    # check if empty
    with pytest.raises(IOError) as excinfo:
        check_and_create_folder(
            foldername, f"Folder '{foldername.stem}'", error_if_empty=True
        )
    assert (f"Folder '{foldername.stem}' is empty.") in str(excinfo.value)



import configparser
import logging
import platform
import shutil
from pathlib import Path

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def format_forced_routing(gdf, minmax, resolution, trajectory_routing=False):
    """Format forced routing LineString geometry in a dataframe writable to the ini-file

    Parameters
    ----------
    gdf: GeoPandas.GeoDataFrame
        Should contain geometry-proprety with LineStrings
    minmax: list
        [x_left, y_lower, x_right, y_upper]. The **first** entry is **x_left**, the
        **second** entry is **y_lower**, the **third** entry is **x_right** and the
        **fourth** entry is **y_upper**.
    resolution: int
        In m
    trajectory_routing: bool, default False
        Use routing via trajectory (True) or via jump (False)

    Returns
    -------
    df: pandas.DataFrame
        Each row is one routing element, with columns:

        - fromX (float): source X-coordinate
        - fromY (float): source Y-coordinate
        - toX (float): target X-coordinate
        - toY (float): target Y-coordinate
        - fromcol (int): source column
        - fromrow (int): source row
        - tocol (int): target column
        - torow (int): target row
    """

    lst_df = []
    for it in gdf.geometry:
        lst_df.append(
            reformat_LineString_to_source_targetf(
                it.coords, trajectory_routing=trajectory_routing
            )
        )
    df = pd.concat(lst_df, ignore_index=True)
    df["fromcol"] = (np.floor((df["fromX"] - minmax[0]) / resolution) + 1).astype(int)
    df["fromrow"] = (np.floor((minmax[3] - df["fromY"]) / resolution) + 1).astype(int)
    df["tocol"] = (np.floor((df["toX"] - minmax[0]) / resolution) + 1).astype(int)
    df["torow"] = (np.floor(minmax[3] - df["toY"]) / resolution + 1).astype(int)
    cond = (df["fromcol"] == df["tocol"]) & (df["fromrow"] == df["torow"])
    df = df.loc[~cond]

    return df


def reformat_LineString_to_source_targetf(coordinates, trajectory_routing=False):
    """Reformat single LineString-routing to source/target dataframe format

    The routing can be organised via a trajectory or a jump:

    - trajectory: the routing follows the sequential coordinates of the elements in the
      LineString.
    - jump: the routing goes from the start to the end of the LineString.

    Parameters
    ----------
    coordinates: list
        List of tuples (x,y)
    trajectory_routing: bool, default False
        Use routing via trajectory (True) or via jump (False)

    Returns
    -------
    df: pandas.DataFrame
        Each row is one routing element, with columns:

        - fromX (float): source X-coordinate
        - fromY (float): source Y-coordinate
        - toX (float): target X-coordinate
        - toY (float): target Y-coordinate
    """
    df = pd.DataFrame()
    if trajectory_routing:
        df[["fromX", "fromY"]] = [(x, y) for x, y in coordinates]
        df["toX"] = np.nan
        df["toY"] = np.nan
        df["toX"].iloc[0:-1] = df["fromX"].iloc[1:]
        df["toY"].iloc[0:-1] = df["fromY"].iloc[1:]
        df = df.iloc[:-1]
    else:
        df = pd.DataFrame(columns=["fromX", "fromY", "toX", "toY"], index=[0])
        df.loc[0, ["fromX", "fromY"]] = coordinates[0]
        df.loc[0, ["toX", "toY"]] = coordinates[-1]

    return df


def zip_folder(folder2zip, outfile=None):
    """Zip a folder

    Parameters
    ----------
    folder2zip: str
        File path of the folder to be zipped
    outfile: str
        File path and name of the output zip-file (optional). If not given the zip
        file will have the same name as the folder and it will be stored in
        the same place.
    """
    if outfile is None:
        outfile = folder2zip
    shutil.make_archive(outfile, "zip", folder2zip)


def get_item_from_ini(ini, section, option, dtype):
    """Gets an item from an option in a certain section of an ini-file

    Parameters
    ----------
    ini: pathlib.Path
        File path of the ini-file
    section: str
        Name of the desired section
    option: str
        Name of the desired option
    dtype: str
        Type of parameter to be read (get (string), int, float or bool).
        If another string is giver to this parameter, a
        :class:`~pywatemsedem.CNWS.CNWSException` is raised.

    Returns
    -------
    str or int or float or bool or None
        Settings value of ini-file. If the option or section does not exist in
        the ini-file or the value of the section-option None is returned.
    """

    Cfg = configparser.ConfigParser()
    if ini.exists():
        Cfg.read(ini)
        try:
            if dtype == "get":
                a = Cfg.get(section, option)
            elif dtype == "int":
                a = Cfg.getint(section, option)
            elif dtype == "float":
                a = Cfg.getfloat(section, option)
            elif dtype == "bool":
                a = Cfg.getboolean(section, option)
            else:
                raise TypeError("not a correct Type")
        except configparser.NoOptionError:
            msg = f"Option {option} does not exist in ini-file (section {section})"
            logger.warning(msg)
            return None
        except configparser.NoSectionError:
            msg = f"Section {section} does not exist in ini-file"
            logger.warning(msg)
            return None
        else:
            return a
    else:
        raise FileNotFoundError(f"{ini} does not exist")


def extract_tags_from_template_file(template):
    """Extract tags (scenario, catchment, year from the WaTEM/SEDEM perceelskaart template
    file.


    Parameters
    ----------
    template: pathlib.Path or str
        File path to CN-WS perceelskaart template file.

    Returns
    -------
    catchment_name: str
        Catchment name
    scenario: str
        Scenario number (in string!)
    year: int
        Simulation year.
    valid: bool
        Tag that indicate whether template format agrees with defined template format.

    Note
    ----
    The template format should be

    """
    # take first template file and drop extension
    template_, _, _ = Path(template).name.partition(".")

    # split template name on "_"
    template_ = template_.split("_")

    # valid if template conveys with define template format
    if len(template_) != 4:
        msg = (
            f"Template {template} does not convey with defined template format "
            "`fname_%year_%catchmentname_s%scenariolabel.rst`."
        )
        logger.warning(msg)
        valid = False
    else:
        valid = True

    # get catchment name which is on the third position (index 2)
    catchment_name = template_[2]
    # get scenario number (str) from last position and drop "s"
    scenario = template_[-1].replace("s", "")
    # get year (int) from second position (index 1)
    year = int(template_[1])

    return catchment_name, scenario, year, valid


def get_df_area_unique_values_array(arr, resolution):
    """Calculate the total area for every value in an array

    The area for a value is calculated as the amount of cells with
    a certain value * cellsize * cellsize.

    Parameters
    ----------
    arr: numpy.ndarray
        Input array
    resolution: int
        The spatial resolution of the raster

    Returns
    -------
    df: pandas.DataFrame
        With two columns:

        - *values* (float): all values within the raster
        - *area* (float): the area of every value in the raster

    """
    vals, counts = np.unique(arr, return_counts=True)
    areas = np.multiply(counts, resolution**2)
    df = pd.DataFrame()
    df["values"] = vals
    df["area"] = areas
    return df


def check_cn_ws_binary(cn_ws_binary=None, fixed_name_cmd="cn_ws"):
    """
    Check if the compiled CN_WS program is known in the environment,
    or if the path reference exists

    Parameters
    ----------
    cn_ws_binary: str, default None
        File Path of compiled cn_ws Pascal code. If None, check environment.
    fixed_name_cmd: str, default 'cn_ws'
        Name of exe/program.

    Returns
    -------
    cn_ws_exe: str
        File path of compiled cn_ws Pascal code.
    """
    if cn_ws_binary is None:
        cn_ws_defined_in_path = shutil.which(fixed_name_cmd)
        if cn_ws_defined_in_path is None:
            try_location = Path.home() / "GitHub" / "cn_ws" / "cn_ws" / "cn_ws"
            if try_location.exists():
                cn_ws_binary = try_location
            else:
                msg = (
                    "CN_WSmodel exe/program not found in environment, make sure you "
                    "have defined a compiled version of cn_ws in your "
                    "environment correctly!"
                )
                raise FileNotFoundError(msg)
        else:
            cn_ws_binary = fixed_name_cmd
    else:
        if (platform.system() == "Windows") and (Path(cn_ws_binary).suffix is None):
            cn_ws_binary = cn_ws_binary + ".exe"
        if not Path(cn_ws_binary).exists():
            msg = (
                f"The custom-path defined for the CN_WSmodel exe/program "
                f"{cn_ws_binary} is incorrect, please check you refer "
                f"to a compiled CN_WSmodel."
            )
            raise FileNotFoundError(msg)

    return cn_ws_binary


def check_courant_criterium(velocity, time_step, resolution, factor=0.75):
    """Check the courant criterium, see
    https://en.wikipedia.org/wiki/Courant%E2%80%93Friedrichs%E2%80%93Lewy_condition

    Parameters
    ----------
    velocity: float
        Stream velocity (m/s)
    time_step: float
        Time step of calculations (s)
    resolution: float
        Grid resolution (m)
    factor: float
        Correction factor to avoid that CN-WS Pascal does crash if the distance water
        covers in a timestep that is equal to +-0.75 times the pixel resolution
        (temporary fix).

    Returns
    -------
    time_step_posterior: float
        Posterior time step of calculations (s)
    """
    dist_per_calc_step = velocity * time_step
    if dist_per_calc_step > factor * resolution:
        time_step_posterior = int(np.floor(factor * resolution / velocity))
        msg = (
            f"Time step of {time_step} seconds too large for a"
            f"stream velocity of {velocity} "
            f"m/s (Courant criterium) lowering to resolution divided by "
            f"the velocity ({time_step_posterior})."
        )
        logger.warning(msg)
    else:
        time_step_posterior = time_step
    return time_step_posterior



import configparser
import datetime
import logging
import re
import warnings

from pywatemsedem.templates import InputFileName

logger = logging.getLogger(__name__)

inputfilename = InputFileName()


class IniFile:
    """Create WaTEM/SEDEM ini-file

    Parameters
    ----------
    choices: pywatemsedem.pywatemsedem.chocies.UserChoices
        See :class:`pywatemsedem.pywatemsedem.userchoices.UserChoices`
    version: str
        Model version
    inputfolder: pathlib.Path
        WaTEM/SEDEM inputfolder
    outputfolder: pathlib.Path
        WaTEM/SEDEM outputfolder

    Examples
    --------
    >>> choices = Choices(...)
    >>> forced_routing = ...
    >>> buffers = ...
    >>> inifile = IniFile(choices,"WS","modelinput","modeloutput")
    >>> inifile.add_model_information()
    >>> inifile.add_working_directories()
    >>> inifile.add_files()
    >>> inifile.add_user_choices()
    >>> inifile.add_output_maps()
    >>> inifile.add_variables(buffers, forced_routing, river_underground)
    >>> inifile.write("modelinput/ini_file.ini")
    """

    def __init__(self, choices, version, inputfolder, outputfolder):
        """Creates an ini-file for the scenario"""

        self.cnwsinput_folder = inputfolder
        self.cnwsoutput_folder = outputfolder
        self.choices = choices
        self.version = version
        self.cfg = configparser.ConfigParser()

    def write(self, ini):
        """Write ini file to disk"""
        with open(ini, "w") as f:
            self.cfg.write(f)
            f.close()

    def add_model_information(self):
        """Add section model information to config file"""
        self.cfg.add_section("Model information")
        self.cfg.set(
            "Model information",
            "Date",
            datetime.datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
        )

    def add_working_directories(self):
        """Add section working directory to config file"""
        self.cfg.add_section("Working directories")
        self.cfg.set(
            "Working directories", "Input directory", str(self.cnwsinput_folder)
        )
        self.cfg.set(
            "Working directories",
            "Output directory",
            str(self.cnwsoutput_folder),
        )

    def add_files(self):
        """Add section files to config file"""
        self.cfg.add_section("Files")
        self.cfg.set(
            "Files",
            "DTM filename",
            (self.cnwsinput_folder / inputfilename.dtm_file).name,
        )
        self.cfg.set(
            "Files",
            "P factor map filename",
            (self.cnwsinput_folder / inputfilename.pfactor_file).name,
        )
        self.cfg.set(
            "Files",
            "shapefile catchment",
            (self.cnwsinput_folder / inputfilename.mask_file).name,
        )

        self.cfg.set(
            "Files",
            "Parcel filename",
            (self.cnwsinput_folder / inputfilename.parcelmosaic_file).name,
        )
        if self.version != "Only Routing":
            self.cfg.set(
                "Files",
                "C factor map filename",
                (self.cnwsinput_folder / inputfilename.cfactor_file).name,
            )
        else:
            msg = "Ini-file for version 'Only Routing' not implemented."
            raise IOError(msg)

    def add_user_choices(self):
        """Add section user choices to config file"""
        self.cfg.add_section("User Choices")
        self.cfg.set(
            "User Choices", "Max kernel", str(self.choices.dict_variables["Max kernel"])
        )
        self.cfg.set(
            "User Choices",
            "Max kernel river",
            str(self.choices.dict_variables["Max kernel river"]),
        )

    def add_output_maps(self):
        """Add section outputs config file"""
        self.cfg.add_section("Output maps")
        self.cfg.set(
            "Output maps", "Write aspect", str(self.choices.dict_output["Write aspect"])
        )
        self.cfg.set(
            "Output maps",
            "Write LS factor",
            str(self.choices.dict_output["Write LS factor"]),
        )
        self.cfg.set(
            "Output maps",
            "Write rainfall excess",
            str(self.choices.dict_output["Write rainfall excess"]),
        )
        self.cfg.set(
            "Output maps", "Write RUSLE", str(self.choices.dict_output["Write RUSLE"])
        )
        self.cfg.set(
            "Output maps",
            "Write sediment export",
            str(self.choices.dict_output["Write sediment export"]),
        )
        self.cfg.set(
            "Output maps", "Write slope", str(self.choices.dict_output["Write slope"])
        )
        self.cfg.set(
            "Output maps",
            "Write total runoff",
            str(self.choices.dict_output["Write total runoff"]),
        )
        self.cfg.set(
            "Output maps",
            "Write upstream area",
            str(self.choices.dict_output["Write upstream area"]),
        )
        self.cfg.set(
            "Output maps",
            "Write water erosion",
            str(self.choices.dict_output["Write water erosion"]),
        )
        self.cfg.set(
            "Output maps",
            "Write routing table",
            str(self.choices.dict_output["Write routing table"]),
        )

        self.cfg.set(
            "Output maps",
            "Write routing column/row",
            str(self.choices.dict_output["Write routing table"]),
        )

    def add_variables(self, buffers, forced_routing, river_underground):
        """Add section variables to config file

        Parameters
        ----------
        buffers: pandas.DataFrame
            Holding buffer properties, with columns,

            - *buffercap* (float): Volume
            - *hdam* (float): Height dam
            - *hknijp^* (float): Height opening
            - *dknijp* (float): Opening area
            - *qcoef* (float): Discharge coefficient
            - *boverl* (float): Width dam
            - *eff* (float): Trapping efficiency
            - *BUF_EXID* (float): Extension ID
            - *BUF_ID* (float): Buffer id

        forced_routing: pandas.DataFrame
            Forced routing table with columns, see :ref:`here <watemsedem:forcedroutingdata>`

            - *NR* (int):
            - *from col* (int): source pixel column
            - *from row* (int): source pixel row
            - *target col* (int): target pixel column
            - *target row* (int): target pixel row

        river_underground: pandas.DataFrame
            River underground routing table with columns, see
            :ref:`here <watemsedem:forcedroutingdata>`

            - *NR* (int):
            - *from col* (int): source pixel column
            - *from row* (int): source pixel row
            - *target col* (int): target pixel column
            - *target row* (int): target pixel row
        """
        self.cfg.add_section("Variables")
        self.cfg.set(
            "Variables",
            "Parcel connectivity grasstrips",
            str(int(self.choices.dict_variables["Parcel connectivity grasstrips"])),
        )
        self.cfg.set(
            "Variables",
            "Parcel connectivity cropland",
            str(int(self.choices.dict_variables["Parcel connectivity cropland"])),
        )
        self.cfg.set(
            "Variables",
            "Parcel connectivity forest",
            str(int(self.choices.dict_variables["Parcel connectivity forest"])),
        )
        self.cfg.set(
            "Variables",
            "Parcel trapping efficiency cropland",
            str(
                int(self.choices.dict_variables["Parcel trapping efficiency cropland"])
            ),
        )
        self.cfg.set(
            "Variables",
            "Parcel trapping efficiency forest",
            str(int(self.choices.dict_variables["Parcel trapping efficiency forest"])),
        )
        self.cfg.set(
            "Variables",
            "Parcel trapping efficiency pasture",
            str(int(self.choices.dict_variables["Parcel trapping efficiency pasture"])),
        )
        self.cfg.set(
            "Variables",
            "LS correction",
            str(self.choices.dict_variables["LS correction"]),
        )

        self.set_cfg_model_version()

        # output per rivier segment?
        self.cfg.set(
            "User Choices",
            "Output per river segment",
            str(self.choices.dict_output["Output per river segment"]),
        )
        # if self.choices.dict_output["Output per river segment"] == 1:
        self.cfg.set(
            "Files",
            "River segment filename",
            (self.cnwsinput_folder / inputfilename.segments_file).name,
        )

        # manual outlet
        self.cfg.set(
            "User Choices",
            "Manual outlet selection",
            str(self.choices.dict_model_options["Manual outlet selection"]),
        )
        if self.choices.dict_model_options["Manual outlet selection"] == 1:
            self.cfg.set(
                "Files",
                "Outlet map filename",
                (self.cnwsinput_folder / inputfilename.outlet_file).name,
            )

        self.set_cfg_only_routing()

        # Grachten gebruiken?
        self.cfg.set(
            "User Choices",
            "Include ditches",
            str(self.choices.dict_ecm_options["Include ditches"]),
        )
        if self.choices.dict_ecm_options["Include ditches"] == 1:
            self.cfg.set(
                "Files",
                "Ditch map filename",
                (self.cnwsinput_folder / inputfilename.ditches_files).name,
            )

        # Geleidende dammen gebruiken?
        self.cfg.set(
            "User Choices",
            "Include dams",
            str(self.choices.dict_ecm_options["Include dams"]),
        )
        if self.choices.dict_ecm_options["Include dams"] == 1:
            self.cfg.set(
                "Files",
                "Dam map filename",
                (self.cnwsinput_folder / inputfilename.conductivedams_file).name,
            )

        # Using sewers
        self.cfg.set(
            "User Choices",
            "Include sewers",
            str(self.choices.dict_model_options["Include sewers"]),
        )
        if self.choices.dict_model_options["Include sewers"] == 1:
            self.cfg.set(
                "Variables",
                "Sewer exit",
                str(self.choices.dict_variables["Sewer exit"]),
            )
            self.cfg.set(
                "Files",
                "Sewer map filename",
                (self.cnwsinput_folder / inputfilename.endpoints_file).name,
            )

        # Using buffers in model?
        self.cfg.set(
            "User Choices",
            "Include buffers",
            str(self.choices.dict_ecm_options["Include buffers"]),
        )
        if self.choices.dict_ecm_options["Include buffers"] == 1:
            buffers = [] if buffers is None else buffers
            self.cfg.set(
                "Files",
                "Buffer map filename",
                (self.cnwsinput_folder / inputfilename.buffers_file).name,
            )
            self.cfg.set("Variables", "Number of buffers", str(len(buffers)))
            if len(buffers) != 0:
                for row in buffers.iterrows():
                    sectie = f"Buffer {row[1]['buf_id']}"
                    self.cfg.add_section(sectie)
                    self.cfg.set(sectie, "Volume", f"{row[1]['buffercap']}")
                    self.cfg.set(sectie, "Height dam", f"{row[1]['hdam']}")
                    self.cfg.set(sectie, "Height opening", f"{row[1]['hknijp']}")
                    self.cfg.set(sectie, "Opening area", f"{row[1]['dknijp']}")
                    self.cfg.set(sectie, "Discharge coefficient", f" {row[1]['qcoef']}")
                    self.cfg.set(sectie, "Width dam", f"{row[1]['boverl']}")
                    self.cfg.set(sectie, "Trapping efficiency", f"{row[1]['eff']}")
                    self.cfg.set(sectie, "Extension ID", f"{row[1]['buf_exid']}")

        # force routing
        self.cfg.set(
            "User Choices",
            "Force Routing",
            str(self.choices.dict_model_options["Force Routing"]),
        )
        if self.choices.dict_model_options["Force Routing"] == 1:
            forced_routing = [] if forced_routing is None else forced_routing
            river_underground = [] if river_underground is None else river_underground
            n_fr = len(forced_routing)
            n_ru = len(river_underground)
            n = n_ru + n_fr
            if n > 0:
                self.cfg.set(
                    "Variables",
                    "Number of Forced Routing",
                    str(n),
                )
            if n_fr > 0:
                for nr, row in enumerate(forced_routing.itertuples()):
                    sectie = f"Forced Routing {nr+1}"
                    self.cfg.add_section(sectie)
                    self.cfg.set(sectie, "from col", f"{int(row.fromcol)}")
                    self.cfg.set(sectie, "from row", f"{int(row.fromrow)}")
                    self.cfg.set(sectie, "target col", f"{int(row.tocol)}")
                    self.cfg.set(sectie, "target row", f"{int(row.torow)}")
            elif n_ru > 0:
                for nr, row in enumerate(river_underground.itertuples()):
                    sectie = f"Forced Routing {nr + 1}"
                    self.cfg.add_section(sectie)
                    self.cfg.set(sectie, "from col", f"{int(row.fromcol)}")
                    self.cfg.set(sectie, "from row", f"{int(row.fromrow)}")
                    self.cfg.set(sectie, "target col", f"{int(row.tocol)}")
                    self.cfg.set(sectie, "target row", f"{int(row.torow)}")
            else:
                msg = (
                    "'Force Routing' in 'dict_model_options' is equal to 1, but no "
                    "forced routing or underground or river vectors defined."
                )
                warnings.warn(msg)
        # river routing
        if self.choices.dict_model_options["River Routing"] == 1:
            self.cfg.set(
                "User Choices",
                "river routing",
                str(self.choices.dict_model_options["River Routing"]),
            )
            self.cfg.set(
                "Files",
                "adjectant segments",
                (self.cnwsinput_folder / inputfilename.adjacentedges_file).name,
            )
            self.cfg.set(
                "Files",
                "upstream segments",
                (self.cnwsinput_folder / inputfilename.upedges_file).name,
            )
            self.cfg.set(
                "Files",
                "river routing filename",
                (self.cnwsinput_folder / inputfilename.routing_file).name,
            )

        # Advanced settings
        if "L model" in self.choices.dict_model_options.keys():
            self.cfg.set(
                "User Choices", "L model", self.choices.dict_model_options["L model"]
            )
        if "S model" in self.choices.dict_model_options.keys():
            self.cfg.set(
                "User Choices", "S model", self.choices.dict_model_options["S model"]
            )
        if "TC model" in self.choices.dict_model_options.keys():
            self.cfg.set(
                "User Choices", "TC model", self.choices.dict_model_options["TC model"]
            )

        self.cfg.set(
            "User Choices",
            "Adjusted Slope",
            str(self.choices.dict_model_options["Adjusted Slope"]),
        )
        self.cfg.set(
            "User Choices",
            "Buffer reduce Area",
            str(self.choices.dict_model_options["Buffer reduce Area"]),
        )

    def set_cfg_model_version(self):
        """Add model version to self.cfg."""
        if self.version == "CN-WS":
            self.cfg.set("User Choices", "Only WS", "0")
            self.cfg.set(
                "Variables", "Alpha", str(self.choices.dict_variables["Alpha"])
            )
            self.cfg.set("Variables", "Beta", str(self.choices.dict_variables["Beta"]))
            self.cfg.set(
                "Variables",
                "Stream velocity",
                str(self.choices.dict_variables["Stream velocity"]),
            )
            self.cfg.set(
                "Files",
                "CN map filename",
                (self.cnwsinput_folder / inputfilename.cn_file).name,
            )

        else:  # WS-model
            self.cfg.set("User Choices", "Only WS", "1")
            if self.version == "Only Routing":
                self.cfg.set("User Choices", "Only Routing", "1")

    def set_cfg_only_routing(self):
        """Add variables for only routing version of WaTEM/SEDEM."""
        if self.version != "Only Routing":
            self.cfg.set(
                "Files",
                "K factor filename",
                (self.cnwsinput_folder / inputfilename.kfactor_file).name,
            )
            self.cfg.set(
                "Variables",
                "Bulk density",
                str(int(self.choices.dict_variables["Bulk density"])),
            )

            # R-factor of regenvalfile?
            self.cfg.set(
                "Variables",
                "R factor",
                str(self.choices.dict_variables["R factor"]),
            )
            self.cfg.set("Variables", "Endtime model", "0")  # Aan te passen in model?
            if self.version == "CN-WS":
                self.cfg.set(
                    "Files",
                    "Rainfall filename",
                    "Rainfall.txt",  # TODO: define in templates
                )
                self.cfg.set(
                    "Variables",
                    "5-day antecedent rainfall",
                    str(self.choices.dict_variables["5-day antecedent rainfall"]),
                )
                self.cfg.set(
                    "Variables",
                    "Desired timestep for model",
                    str(self.choices.dict_variables["Desired timestep for model"]),
                )
                self.cfg.set(
                    "Variables",
                    "Endtime model",
                    str(self.choices.dict_variables["Endtime model"]),
                )
                if self.choices.dict_model_options["Convert output"] == 1:
                    self.cfg.set(
                        "User Choices",
                        "Convert output",
                        str(self.choices.dict_model_options["Convert output"]),
                    )
                    self.cfg.set(
                        "Variables",
                        "Final timestep output",
                        str(self.choices.dict_variables["Final timestep output"]),
                    )

            # KTC and calibration KTC
            if self.choices.dict_model_options["Calibrate"] == 1:
                self.cfg.add_section("Calibration")
                self.cfg.set(
                    "Calibration",
                    "Calibrate",
                    str(self.choices.dict_model_options["Calibrate"]),
                )
                self.cfg.set(
                    "Calibration",
                    "KTcHigh_lower",
                    str(self.choices.dict_variables["KTcHigh_lower"]),
                )
                self.cfg.set(
                    "Calibration",
                    "KTcHigh_upper",
                    str(self.choices.dict_variables["KTcHigh_upper"]),
                )
                self.cfg.set(
                    "Calibration",
                    "KTcLow_lower",
                    str(self.choices.dict_variables["KTcLow_lower"]),
                )
                self.cfg.set(
                    "Calibration",
                    "KTcLow_upper",
                    str(self.choices.dict_variables["KTcLow_upper"]),
                )
                self.cfg.set(
                    "Calibration", "steps", str(self.choices.dict_variables["steps"])
                )
                self.cfg.set(
                    "Variables",
                    "ktc limit",
                    str(self.choices.dict_variables["ktc limit"]),
                )
            elif self.choices.dict_model_options["UserProvidedKTC"] == 1:
                self.cfg.set("User Choices", "Create ktc map", "0")
                self.cfg.set(
                    "Files",
                    "ktc map filename",
                    (self.cnwsinput_folder / inputfilename.ktc_file).name,
                )

            else:
                self.cfg.set("User Choices", "Create ktc map", "1")
                self.cfg.set(
                    "Variables", "ktc low", str(self.choices.dict_variables["ktc low"])
                )
                self.cfg.set(
                    "Variables",
                    "ktc high",
                    str(self.choices.dict_variables["ktc high"]),
                )
                self.cfg.set(
                    "Variables",
                    "ktc limit",
                    str(self.choices.dict_variables["ktc limit"]),
                )

            # KTIL
            self.cfg.set(
                "User Choices",
                "Calculate Tillage Erosion",
                str(self.choices.dict_model_options["Calculate Tillage Erosion"]),
            )
            if self.choices.dict_model_options["Calculate Tillage Erosion"] == 1:
                self.cfg.set(
                    "User Choices",
                    "Create ktil map",
                    str(self.choices.dict_model_options["Create ktil map"]),
                )
                if self.choices.dict_model_options["Create ktil map"] == 1:
                    self.cfg.set(
                        "Variables",
                        "ktil default",
                        str(self.choices.dict_variables["ktil default"]),
                    )
                    self.cfg.set(
                        "Variables",
                        "ktil threshold",
                        str(self.choices.dict_variables["ktil threshold"]),
                    )

            self.cfg.set(
                "User Choices",
                "Estimate clay content",
                str(self.choices.dict_model_options["Estimate clay content"]),
            )
            if self.choices.dict_model_options["Estimate clay content"] == 1:
                self.cfg.set(
                    "Variables",
                    "Clay content parent material",
                    str(
                        int(self.choices.dict_variables["Clay content parent material"])
                    ),
                )


def open_ini(ini):
    """Open ini file.

    Parameters
    ----------
    ini: pathlib.Path

    Returns
    -------
    content: list
    """
    with open(ini, "r") as file:
        content = file.readlines()
    return content


def modify_field(ini, section, key, value):
    """Add one specific key in ini-file

    Loads, writes value to key (in section) and closes file

    Parameters
    ----------
    ini: pathlib.Path
        File name
    section: str
        Section name, i.e. Model information, Working directories, Files, User Choices,
        Output , Variables, Calibration, Buffer $id and Forced Routing $id.
    key: str
        Key
    new_value: str
        New value
    ini_updated: pathlib.Path, default None
        New file to write to
    """
    # Init
    content = open_ini(ini)
    section_pattern = re.compile(rf"^\[{re.escape(section)}\]\s*$")
    key_pattern = re.compile(rf"^{re.escape(key)}\s*=")
    in_section = False
    in_key = False

    # Modify
    for i, line in enumerate(content):
        if section_pattern.match(line):
            in_section = True
        elif in_section and key_pattern.match(line):
            content[i] = f"{key} = {value}\n"
            in_key = True
            break

    # Warning
    if not in_section:
        msg = (
            f"Cannot modify '{value}' to '{key}'. Section '{section}' does not exist "
            f"in '{ini}'."
        )
        raise KeyError(msg)
    if not in_key:
        msg = (
            f"Cannot modify '{value}' to '{key}'. Key '{key}' does not exist in "
            f"'{ini}'."
        )
        raise KeyError(msg)

    # Write the modified content back to the file
    with open(ini, "w") as file:
        msg = f"Modifying '{value}' to '{key}' in file '{ini}'"
        logger.info(msg)
        file.writelines(content)


def add_field(ini, section, key, value):
    """Add one specific key in ini-file

    Loads, writes value to key (in section) and closes file

    Parameters
    ----------
    ini: pathlib.Path
        File name
    section: str
        Section name, i.e. Model information, Working directories, Files, User Choices,
        Output , Variables, Calibration, Buffer $id and Forced Routing $id.
    key: str
        Key
    value: str
        New value

    """
    # Init
    content = open_ini(ini)
    section_pattern = re.compile(rf"^\[{re.escape(section)}\]\s*$")
    key_pattern = re.compile(rf"^{re.escape(key)}\s*=")
    in_section = False

    # Modify
    for i, line in enumerate(content):
        if section_pattern.match(line):
            in_section = True
        elif in_section and line.startswith("["):
            content_before_insertion = content[: i - 1]
            content_after_insertion = content[i:]
            content = (
                content_before_insertion
                + [f"{key} = {value} \n \n"]
                + content_after_insertion
            )
            break
        elif in_section and key_pattern.match(line):
            msg = (
                f"Cannot add '{value}' to '{key}'. Key '{key}' already exist in "
                f"'{ini}'."
            )
            raise KeyError(msg)
        elif in_section and i == (len(content) - 1):
            content = content + [f"{key} = {value}\n"]

    # Warning
    if not in_section:
        msg = (
            f"Cannot add '{value}' to '{key}'. Section '{section}' does not exist "
            f"in '{ini}'."
        )
        raise KeyError(msg)

    # Write the modified content back to the file
    with open(ini, "w") as file:
        msg = f"Adding '{value}' to '{key}' in file '{ini}'"
        logger.info(msg)
        file.writelines(content)



import warnings

import numpy as np


def valid_boundaries(arr, lower=None, upper=None):
    """Checks if values are within specified boundaries

    Parameters
    ---------
    arr: numpy.ndarray
        Input array
    lower: int, float, default None
        Lower boundary
    upper: int, float, default None
        Upper boundary

    Returns
    -------
    True
    """

    if (lower is None) and (upper is None):
        msg = "No lower or upper boundaries are defined, omitting boundary check"
        warnings.warn(msg)

    if upper is not None:
        if not isinstance(upper, (int, float)):
            msg = "Upper boundary should be numeric."
            raise TypeError(msg)
        if np.any(arr > upper):
            raise ValueError(f"Values are higher than upper bound ('{upper}')")
    if lower is not None:
        if not isinstance(lower, (int, float)):
            msg = "Lower boundary should be numeric."
            raise TypeError(msg)
        if np.any(arr < lower):
            raise ValueError(f"Values are lower than lower bound ('{lower}')")

    return True


def valid_values(arr, unique_values):
    """Checks if values in an array are within the set of unique values

    The input array can only have the input values defined in unique_values.

    Parameters
    ----------
    arr: numpy.ndarray
        Input array
    unique_values: list
        Unique values that the raster can have

    Returns
    -------
    True
    """

    if not (set(np.unique(arr)).issubset(set(unique_values))):
        msg = "The array contains values that are not present in unique_values"
        raise ValueError(msg)

    return True


def valid_non_nan(arr):
    """Checks if Nan are present in the array

    Parameters
    ----------
    arr: numpy.ndarray
        Input array

    Returns
    -------
    True
    """

    if np.any(np.isnan(arr)):
        msg = "Input array can not contain nan values"
        raise ValueError(msg)

    return True


def valid_array_type(arr, required_type):
    """Checks if arr is
    1) a numpy.dtype
    2) the desired numpy.dtype

    Parameters
    ----------
    arr: numpy.ndarray
        Input array
    required_type: numpy.dtype
         required datatype.

    Returns
    -------
    True
    """
    try:
        if not (np.issubdtype(required_type, np.number)):
            msg = f"No valid numpy dtype chosen for arr! Change to '{required_type}'"
            raise TypeError(msg)
    except TypeError as e:
        msg = str(e) + ". Use numpy dtype as required type."
        raise TypeError(msg)

    if arr.dtype != required_type:
        msg = f"""numpy.dtype (`np.{arr.dtype}`) is not required type
        ('{required_type}')"""
        raise ValueError(msg)

    return True


def valid_nodata(arr, nodata_value=-9999):
    """Checks if a nodata value is present in arr

    Parameters
    --------
    arr: numpy.ndarray
        Input array
    nodata_value: int, default -9999
                    Value to be interpreted as no data present

    Returns
    -------
    True
    """
    if np.any(arr == nodata_value):
        msg = f"A nodata value ({nodata_value}) is present, check if allowed"
        raise Warning(msg)



import sys

if sys.version_info[:2] >= (3, 8):
    # TODO: Import directly (no need for conditional) when `python_requires = >= 3.8`
    from importlib.metadata import PackageNotFoundError  # pragma: no cover
    from importlib.metadata import version
else:
    from importlib_metadata import PackageNotFoundError  # pragma: no cover
    from importlib_metadata import version

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError



import tempfile

import numpy as np
import pytest
from conftest import geodata

from pywatemsedem.geo.rasterproperties import RasterProperties
from pywatemsedem.geo.rasters import RasterFile, RasterMemory, TemporalRaster
from pywatemsedem.geo.utils import load_raster


def test_rastermemory():
    """Test functionalities of RasterMemory class"""

    # mask a certain part of the raster.
    arr, profile = load_raster(geodata.rst_example)
    arr_mask, profile_mask = load_raster(geodata.rst_mask)

    # rasterproperties
    rp = RasterProperties.from_rasterio(profile)
    raster = RasterMemory(arr, rp)

    # test write
    tiff_temp = tempfile.NamedTemporaryFile(suffix=".tif").name
    assert raster.write(tiff_temp, format="tiff", dtype=np.float64)
    tiff_temp = tempfile.NamedTemporaryFile(suffix=".rst").name
    assert raster.write(tiff_temp, dtype=np.int32)
    assert raster.write(tiff_temp, dtype="int32")
    with pytest.raises(NotImplementedError) as excinfo:
        raster.write(tiff_temp, format="test")
    assert "Format 'test' not implemented in pywatemsedem." in str(excinfo.value)
    # enable mask to be in same nodata
    arr_mask = arr_mask.astype(np.float64)
    # feed wrong format
    with pytest.raises(TypeError) as excinfo:
        raster.write(tiff_temp, format="tiff")
    assert "in format 'tiff' with '.rst' extension." in str(excinfo.value)

    assert raster.arr.shape[0] == 286
    assert raster.arr.shape[1] == 458
    raster.mask(arr_mask)
    assert rp.nodata in raster.arr

    # mask array has wrong size
    arr_mask_w = np.vstack([arr_mask, arr_mask[-1, :]])
    with pytest.raises(IOError) as excinfo:
        raster = RasterMemory(arr, rp)
        raster.mask(arr_mask_w)
    assert "Mask array has different size from input array" in str(excinfo.value)

    # wrong values in mask
    arr_mask[arr_mask == 1] = 2
    arr_mask[arr_mask == rp.nodata] = 100
    with pytest.raises(ValueError) as excinfo:
        raster = RasterMemory(arr, rp)
        raster.mask(arr_mask)
    assert "Mask array should have values 1 (no mask) and" in str(excinfo.value)

    # mask leads to a nodata raster
    arr_mask = np.zeros(raster.arr.shape, dtype=np.float32)
    arr_mask[0, 0] = 1
    arr = raster.arr
    arr[0, 0] = rp.nodata
    with pytest.raises(ValueError) as excinfo:
        raster = RasterMemory(arr, rp)
        raster.mask(arr_mask)
    assert (
        "Array after masking has only nodata values/is empty. Please check your "
        "mask/input array." in str(excinfo.value)
    )

    # feed wrong number of dimensions
    arr = raster.arr.flatten()  # flatten to make raster 2 to 1-D.
    with pytest.raises(ValueError) as excinfo:
        RasterMemory(arr, rp)
    assert "Dimensionality of input raster array should be larger than 1" in str(
        excinfo.value
    )


def test_rasterfile():
    """Test functionalities of RasterFile class"""

    # clip
    rp = RasterProperties([162300, 165760, 167560, 169520], 20, -9999, 31370)
    raster = RasterFile(geodata.rst_example, rp=rp)
    assert raster.arr.shape[0] == 188
    assert raster.arr.shape[1] == 263

    # clip with extent outside rasters extent
    rp = RasterProperties([230, 760, 560, 1000], 20, -9999, 31370)
    with pytest.raises(Exception) as excinfo:
        RasterFile(geodata.rst_example, rp=rp)
    assert "Clipped output raster is empty" in str(excinfo.value)

    # do not clip
    raster = RasterFile(geodata.rst_example)
    assert raster.arr.shape[0] == 286
    assert raster.arr.shape[1] == 458

    # check for wrong coordinate system
    rp = RasterProperties([162300, 165760, 167560, 169520], 20, -9999, 3395)
    with pytest.raises(Exception) as excinfo:
        RasterFile(geodata.rst_example, rp=rp)
    assert "should be same as epsg of input raster properties " in str(excinfo.value)


def test_temporalraster():
    """Test functionalities of RasterProperties class"""

    # load raster
    arr, profile = load_raster(geodata.rst_example)
    rp = RasterProperties.from_rasterio(profile)

    # fail when generate with 2D-array
    with pytest.raises(ValueError) as excinfo:
        TemporalRaster(arr, rp)
    assert "Dimensionality of temporal raster array should be equal to 3." in str(
        excinfo.value
    )

    # fail when number of write outputs is
    arr = np.dstack([arr, arr])
    tr = TemporalRaster(arr, rp)
    tiff_temp1 = tempfile.NamedTemporaryFile(suffix=".tif").name
    tiff_temp2 = tempfile.NamedTemporaryFile(suffix=".tif").name

    with pytest.raises(ValueError) as excinfo:
        tr.write([tiff_temp1])
    assert "should be equal to number of arrays in the third dimension" in str(
        excinfo.value
    )

    assert tr.write([tiff_temp1, tiff_temp2], format="tiff")



import geopandas as gpd
import pytest
from conftest import geodata

from pywatemsedem.geo.utils import get_geometry_type
from pywatemsedem.geo.vectors import VectorFile, VectorMemory


def test_vector_file():
    """Test functionalities of VectorFile class"""

    # clip
    vector = VectorFile(geodata.vct_example, vct_clip=geodata.catchment)
    assert len(vector.geodata) == 18


def test_vector_memory():
    """Test functionalities of VectorMemory class"""
    # load
    gdf = gpd.read_file(geodata.vct_example)
    geometry_type = get_geometry_type(geodata.vct_example)

    # unknown required geometry type
    req_geometry_type = "Line"
    with pytest.raises(TypeError) as excinfo:
        VectorMemory(gdf, geometry_type, req_geometry_type)
    assert (
        f"Required geometry item type '{req_geometry_type}' not known to pywatemsedem. "
        f"Please select 'LineString or Polygon or Point'" in str(excinfo.value)
    )

    # wrong required geometry type
    req_geometry_type = "Polygon"
    with pytest.raises(TypeError) as excinfo:
        VectorMemory(gdf, geometry_type, req_geometry_type)
    assert (
        f"Input vector should have geometry item type '{req_geometry_type}', not "
        f"'{geometry_type}'." in str(excinfo.value)
    )

    # empty dataframe
    req_geometry_type = "LineString"
    with pytest.raises(ValueError) as excinfo:
        VectorMemory(gdf[gdf["NR"] == -1], geometry_type, req_geometry_type)
    assert (
        "Input vector is empty. If you wish to return an empty vector, please use "
        "'allow_empty'" in str(excinfo.value)
    )

    # correct input
    vector = VectorMemory(gdf, geometry_type, req_geometry_type)
    assert len(vector.geodata) == 19



"""References to all data used for testing"""
# -*- coding: utf-8 -*-

from pathlib import Path

import pandas as pd
import pytest
from dotenv import find_dotenv, load_dotenv

from pywatemsedem.catchment import Catchment
from pywatemsedem.scenario import Scenario
from pywatemsedem.userchoices import UserChoices

# Load environmental variables
## load dotenv
load_dotenv(find_dotenv())

# subfolder data
folder_core = Path("tests/data")
folder_geo = Path("tests/geo/data")
folder_io = Path("tests/io/data")
userchoices_file = folder_core / "userchoices.ini"
ini_file = folder_io / "model_in" / "ini.ini"


# class rainfall_data:
#     """Rainfall data for 'P06_014', first five days january 2018"""
#
#     rainfall = folder_flanders / "rainfall_nukerke.csv"


class io_data:

    composite_landuse = folder_io / "model_in" / "perceelskaart.rst"


class geodata:
    """Data class for geo test data"""

    rst_mask = folder_geo / "mask.tif"
    vct_mak = folder_geo / "mask.shp"
    catchment = folder_geo / "catchm_langegracht.shp"
    rst_example = folder_geo / "example_input_raster.tif"
    vct_example = folder_geo / "example_input_vector.shp"


class catchment_data:
    """Data class for catchment tests data"""

    # vectors
    catchment = geodata.catchment
    infrastructure = folder_core / "infrastructure.shp"
    river = folder_core / "river.shp"
    roads = folder_core / "roads.shp"
    water = folder_core / "water.shp"

    # rasters
    dtm = folder_core / "dtm.tif"
    k = folder_core / "K3.tif"
    basemap = folder_core / "basemap_landuse.tif"
    hsg = folder_core / "hydrological_soil_groups.tif"
    mask = folder_geo / "mask.tif"


class scenario_data:
    """Data class for scenario tests data"""

    # vectors
    grass_strips = folder_core / "grass_strips.shp"
    buffers = folder_core / "buffers.shp"
    cropinfo = folder_core / "cropinfo.csv"
    outlets = folder_core / "outlets.shp"
    endpoints = folder_core / "endpoints.shp"
    parcels = folder_core / "parcels.shp"
    force_routing = folder_core / "force_routing.shp"


# class ecm_data_flanders:
#     """Data class for ecm flanders tests data"""
#
#     folder_ecm_model = folder_flanders / "ecm_model"
#
#     vct_bo_fixed = folder_ecm_model / "BO_VLM_2018_EXTERN_aangepast_la72_demer.shp"
#     vct_bo_var = folder_ecm_model / "BO_VAR_TOEGEP_2018_EXTERN_la72_demer.shp"
#     vct_erosiebestrijdingswerken = (
#         folder_ecm_model / "erosiebestrijdingswerken_demer.shp"
#     )
#     vct_erosiebestrijding_line = (
#         folder_ecm_model / "erosiebestrijdingsplan_line_demer.shp"
#     )
#     vct_erosiebestrijding_point = (
#         folder_ecm_model / "erosiebestrijdingsplan_point_demer.shp"
#     )
#     vct_waterline = folder_ecm_model / "topology_rivers_demer.shp"
#     vct_parcels = folder_ecm_model / "percelen_2018_demer.shp"
#     dict_extra_ecm_data = {}
#     dict_extra_ecm_data["buffer"] = folder_ecm_model / "extra_buffer_demer.shp"
#     dict_extra_ecm_data["gras"] = folder_ecm_model / "extra_gras_demer.shp"
#     txt_output_ecm_model = folder_ecm_model / "results_ecm_model_flanders.csv"


class postprocess:
    """Data class for postprocess tests data"""

    folder_postprocess = folder_core / "postprocess"
    rst_prclskrt = folder_postprocess / "perceelskaart_2018_molenbeek_s1.rst"
    rst_buffers = folder_postprocess / "buffers.rst"
    txt_rainfall = folder_postprocess / "cn_synthetic_rainfall_event.txt"

    ## Outputs
    rst_sediin = folder_postprocess / "SediIn_kg.rst"
    rst_sediout = folder_postprocess / "SediOut_kg.rst"
    rst_watereros = folder_postprocess / "WATEREROS (kg per gridcel).rst"
    txt_routing = folder_postprocess / "routing.txt"
    rst_sediexport = folder_postprocess / "SediExport_kg.rst"
    rst_sewerin = folder_postprocess / "sewer_in.rst"

    ## Postprocess
    ### netto erosion
    txt_average_netto_erosion = folder_postprocess / "average_netto_erosion.txt"
    txt_std_dev_netto_erosion = folder_postprocess / "std_dev_netto_erosion.txt"
    txt_sum_netto_erosion = folder_postprocess / "sum_netto_erosion.txt"
    txt_area_parcel = folder_postprocess / "area_parcel.txt"

    ### grass strips
    txt_gras_efficiency = folder_postprocess / "grass_strips_effiency.csv"

    ## Extra data
    rst_rasterized_prc_shp = folder_postprocess / "percelen.tif"

    ### Grass strips id
    rst_grasstrips_id = folder_postprocess / "grasstrips_id.tif"


class grassstripsdata:
    """Data class for grass strips tests data"""

    folder_grass_strips = folder_core / "grass_strips"
    txt_ktc = folder_grass_strips / "ktc_scaling_gras_strips.csv"
    df_ktc = pd.read_csv(txt_ktc)


class calibratedata:
    """Data class for calibration tests data"""

    ### Calibrate
    folder_calibrate = folder_core / "calibrate"
    txt_calibrate = folder_calibrate / "calibration.csv"


class tools:
    """Data class for tools tests data"""

    folder_tools = folder_core / "tools"
    vct_prcln_mstbnk = folder_tools / "mestbank.shp"
    vct_prcln_dlv = folder_tools / "dlv.shp"
    vct_prcln_fltrd_output = folder_tools / "output.shp"
    # Next two layers are used to test self intersection dlv data layers
    vct_input_dlv_preprocess = folder_tools / "input_dlv_preprocess.shp"
    vct_output_dlv_preprocess = folder_tools / "output_dlv_preprocess.shp"


# class cfactormodelflanders:
#
#     folder_cfactor_model = folder_flanders / "cfactor_model"
#     txt_parcels = folder_cfactor_model / "parcels.csv"


@pytest.fixture(scope="session")
def scenario():
    """Initialize catchment

    Save it to cache with request.config.cache once it is made (faster rerun)
    """
    name = "langegracht"
    catchment = Catchment(
        name, catchment_data.catchment, catchment_data.dtm, 20, 31370, -9999
    )
    catchment.kfactor = catchment_data.k
    catchment.landuse = catchment_data.basemap
    catchment.cn = catchment_data.hsg
    catchment.vct_river = catchment_data.river
    catchment.vct_infrastructure_buildings = catchment_data.infrastructure
    catchment.vct_infrastructure_roads = catchment_data.roads
    catchment.vct_water = catchment_data.water

    """Initialize user choices"""
    choices = UserChoices()
    choices.set_ecm_options(userchoices_file)
    choices.set_model_version("WS")
    choices.set_model_options(userchoices_file)
    choices.set_model_variables(userchoices_file)
    choices.set_output(userchoices_file)

    val = Scenario(catchment, 2019, 1, choices)

    return val


# class non_specific_landuse_measures:
#     """Standard scenario parameters and output data for the flanders pipeline.
#
#     We encourage developers to have a look at the standard_scenarios_flanders-module
#     for an explanation of the parameters.
#
#     Used for testing benchmarks for Flanders. The output numbers in this function are
#     hardcoded and checked/adjusted manually.
#
#     This scenario covers the output data for the standard use case for Flanders:
#
#     - No specific landuse: C-factor for agricultural fields is set to 0.37, see
#       :func:`pywatemsedem.flanders.standard_scenarios.apply_specific_landuse`.
#
#     - Following symptom-oriented measures are implemented:
#
#         - 'beheerovereenkomsten'
#         - 'erosiebestrijdingswerken'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Buffers and grass strips are used, source oriented measures not.
#     """
#
#     # input choices
#     specific_landuse = False
#     c_factor_model = None
#     symptom_oriented_packages = ["beheerovereenkomsten", "erosiebestrijdingswerken"]
#     source_oriented_packages = None
#     buffers = True
#     grass_strips = True
#     source_oriented = False
#
#     # output: ton
#     erosion = -17.13
#     deposition = 14.44
#     to_river = 0.92
#     outside_domain = 0.19
#     buffers = 0.10
#     to_ditches = 0.70
#     to_sewers = 0.44
#
#
# class non_specific_landuse_no_measures:
#     """Standard scenario parameters and output data for the flanders pipeline.
#
#     We encourage developers to have a look at the standard_scenarios_flanders-module
#     for an explanation of the parameters.
#
#     Used for testing benchmarks for Flanders. The output numbers in this function are
#     hardcoded and checked/adjusted manually.
#
#     This scenario covers the output data for the standard use case for Flanders:
#
#     - No specific landuse: C-factor for agricultural fields is set to 0.37, see
#       :func:`pywatemsedem.flanders.standard_scenarios.apply_specific_landuse`.
#
#     - No symptom-oriented measures are implemented:
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - No buffers, grass strips, source oriented measures are used.
#     """
#
#     # input choices
#     specific_landuse = False
#     c_factor_model = None
#     symptom_oriented_packages = None
#     source_oriented_packages = None
#     buffers = False
#     grass_strips = False
#     source_oriented = False
#
#     # output ws Molenbeek: ton
#     erosion = -20.12
#     deposition = 15.94
#     to_river = 1.41
#     outside_domain = 0.21
#     buffers = 0.00
#     to_ditches = 1.18
#     to_sewers = 0.78
#
#     # output cnws Nukerke: m3/s ...
#     runoff_count = 3.820000e02
#     runoff_mean = 4.084430e-02
#     runoff_std = 1.101224e-01
#     runoff_min = 2.101948e-47
#     runoff_max = 8.621105e-01
#     runoff_50 = 6.997757e-04
#
#     # ... and g/L
#     concentration_count = 382.000000
#     concentration_mean = 1.546574
#     concentration_std = 1.284577
#     concentration_min = 0.355476
#     concentration_max = 3.121928
#     concentration_50 = 0.565378
#
#
# class specific_landuse_indicator_current_measures:
#     """Standard scenario parameters and output data for the flanders pipeline.
#
#     We encourage developers to have a look at the standard_scenarios_flanders-module
#     for an explanation of the parameters.
#
#     Used for testing benchmarks for Flanders. The output numbers in this function are
#     hardcoded and checked/adjusted manually.
#
#     This scenario covers the output data for the standard use case for Flanders:
#
#     - indicator: C-factors defined in the indicator are used, see
#       :func:`pywatemsedem.flanders.standard_scenarios.apply_specific_landuse`.
#
#     - Following symptom-oriented measures are implemented:
#
#         - 'beheerovereenkomsten'
#         - 'erosiebestrijdingswerken'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Following source-oriented measures are implemented:
#
#         - 'beheerovereenkomsten'
#         - 'randvoorwaarden'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Buffers, grass strips and source-oriented measures are used.
#     """
#
#     # input choices
#     specific_landuse = True
#     c_factor_model = "indicator"
#     symptom_oriented_packages = ["beheerovereenkomsten", "erosiebestrijdingswerken"]
#     source_oriented_packages = ["beheerovereenkomsten", "randvoorwaarden"]
#     buffers = True
#     grass_strips = True
#     source_oriented = True
#
#     # output: ton
#     erosion = -16.40
#     deposition = 13.95
#     to_river = 0.87
#     outside_domain = 0.18
#     buffers = 0.10
#     to_ditches = 0.66
#     to_sewers = 0.36
#
#
# class specific_landuse_indicator_erosiebestrijdingswerken_randvoorwaarden:
#     """Standard scenario parameters and output data for the flanders pipeline.
#
#     We encourage developers to have a look at the standard_scenarios_flanders-module
#     for an explanation of the parameters.
#
#     Used for testing benchmarks for Flanders. The output numbers in this function are
#     hardcoded and checked/adjusted manually.
#
#     This scenario covers the output data for the standard use case for Flanders:
#
#     - indicator: C-factors defined in the indicator are used, see
#       :func:`pywatemsedem.flanders.standard_scenarios.apply_specific_landuse`.
#
#     - Following symptom-oriented measures are implemented:
#
#         - 'erosiebestrijdingswerken'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Following source-oriented measures are implemented:
#
#         - 'randvoorwaarden'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Buffers, grass strips and source-oriented measures are used.
#     """
#
#     # input choices
#     specific_landuse = True
#     c_factor_model = "indicator"
#     symptom_oriented_packages = ["erosiebestrijdingswerken"]
#     source_oriented_packages = ["randvoorwaarden"]
#     buffers = True
#     grass_strips = True
#     source_oriented = True
#
#     # output: ton
#     erosion = -18.67
#     deposition = 14.81
#     to_river = 1.34
#     outside_domain = 0.21
#     buffers = 0.18
#     to_ditches = 1.03
#     to_sewers = 0.59
#
#
# class specific_landuse_indicator_beheerovereenkomsten_randvoorwaarden:
#     """Standard scenario parameters and output data for the flanders pipeline.
#
#     We encourage developers to have a look at the standard_scenarios_flanders-module
#     for an explanation of the parameters.
#
#     Used for testing benchmarks for Flanders. The output numbers in this function are
#     hardcoded and checked/adjusted manually.
#
#     This scenario covers the output data for the standard use case for Flanders:
#
#     - indicator: C-factors defined in the indicator are used, see
#       :func:`pywatemsedem.flanders.standard_scenarios.apply_specific_landuse`.
#
#     - Following symptom-oriented measures are implemented:
#
#         - 'beheerovereenkomsten'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Following source-oriented measures are implemented:
#
#         - 'beheerovereenkomsten'
#         - 'randvoorwaarden'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Buffers, grass strips and source-oriented measures are used.
#     """
#
#     # input choices
#     specific_landuse = True
#     c_factor_model = "indicator"
#     symptom_oriented_packages = ["beheerovereenkomsten"]
#     source_oriented_packages = ["beheerovereenkomsten", "randvoorwaarden"]
#     buffers = True
#     grass_strips = True
#     source_oriented = True
#
#     # output: ton
#     erosion = -17.47
#     deposition = 15.01
#     to_river = 0.91
#     outside_domain = 0.18
#     buffers = 0.00
#     to_ditches = 0.70
#     to_sewers = 0.38
#
#
# class specific_landuse_indicator_beheerovereenkomsten_erosiebestrijdingswerken:
#     """Standard scenario parameters and output data for the flanders pipeline.
#
#     We encourage developers to have a look at the standard_scenarios_flanders-module
#     for an explanation of the parameters.
#
#     Used for testing benchmarks for Flanders. The output numbers in this function are
#     hardcoded and checked/adjusted manually.
#
#     This scenario covers the output data for the standard use case for Flanders:
#
#     - indicator: C-factors defined in the indicator are used, see
#       :func:`pywatemsedem.flanders.standard_scenarios.apply_specific_landuse`.
#
#     - Following symptom-oriented measures are implemented:
#
#         - 'beheerovereenkomsten'
#         - 'erosiebestrijdingswerken'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Following source-oriented measures are implemented:
#
#         - 'beheerovereenkomsten'
#
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Buffers, grass strips and source-oriented measures are used.
#     """
#
#     # input choices
#     specific_landuse = True
#     c_factor_model = "indicator"
#     symptom_oriented_packages = ["beheerovereenkomsten", "erosiebestrijdingswerken"]
#     source_oriented_packages = ["beheerovereenkomsten"]
#     buffers = True
#     grass_strips = True
#     source_oriented = True
#
#     # output: ton
#     erosion = -16.69
#     deposition = 14.12
#     to_river = 0.90
#     outside_domain = 0.19
#     buffers = 0.10
#     to_ditches = 0.68
#     to_sewers = 0.41
#
#
# class specific_landuse_indicator_no_measures:
#     """Standard scenario parameters and output data for the flanders pipeline.
#
#     We encourage developers to have a look at the standard_scenarios_flanders-module
#     for an explanation of the parameters.
#
#     Used for testing benchmarks for Flanders. The output numbers in this function are
#     hardcoded and checked/adjusted manually.
#
#     This scenario covers the output data for the standard use case for Flanders:
#
#     - indicator: C-factors defined in the indicator are used, see
#       :func:`pywatemsedem.flanders.standard_scenarios.apply_specific_landuse`.
#
#     - No symptom-oriented measures.
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - No source-oriented measures are implemented:
#       See
#       :func:`pywatemsedem.flanders.standard_scenarios_flanders.apply_erosion_control_measures`
#
#     - Buffers, grass strips and source-oriented measures are not used.
#     """
#
#     # input choices
#     specific_landuse = True
#     c_factor_model = "indicator"
#     symptom_oriented_packages = None
#     source_oriented_packages = None
#     buffers = False
#     grass_strips = False
#     source_oriented = False
#
#     # output: ton
#     erosion = -19.85
#     deposition = 15.86
#     to_river = 1.41
#     outside_domain = 0.21
#     buffers = 0.00
#     to_ditches = 1.09
#     to_sewers = 0.73
